/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_connection.h 
 * @brief Implementation of the connectionclass.
 * 
 */

#ifndef __GC_CONNECTION_H__
#define __GC_CONNECTION_H__

#include "myx_gc_figure.h"
#include "myx_gc_utilities.h"

//----------------------------------------------------------------------------------------------------------------------

class CGenericCanvas;
class CConnection;
class CConnectionInstance;

typedef enum tagConnectionDirection
{
  GC_DIR_NONE,
  GC_DIR_NORTH,
  GC_DIR_EAST,
  GC_DIR_SOUTH,
  GC_DIR_WEST
} TConnectionDirection;

//----------------- CConnection ----------------------------------------------------------------------------------------

/** A class comprising data for a connection. */
class CConnection: public CGraphicElement
{
private:
  // A private listener class for the figures of a connection.
  class CFigureListener: private CGCListener
  {
    friend class CConnection;
  protected:
    CConnection* connection;
  public:
    virtual void __cdecl onAction(CGCBase* sender, CGCBase* origin, TActionType action) { };
    virtual void __cdecl onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason);
    virtual void __cdecl onDestroy(CGCBase* sender);
    virtual void __cdecl onError(CGCBase* sender, CGCBase* origin, const char* message);
  };

  CFigure* FEnd1;
  CFigure* FEnd2;
  TConnectionDecorations FDecorations;
  CFigureListener FListener;
public:
  CConnection(CGenericCanvas* canvas, CFigure* endPoint1, CFigure* endPoint2);
  virtual ~CConnection(void);

  virtual void __cdecl decorations(const TConnectionDecorations& data);
  virtual TConnectionDecorations __cdecl decorations(void) { return FDecorations; };
  void figureDestroyed(CFigure* figure);
  virtual TGCVariant __cdecl property(const char* name, unsigned int index);
  virtual void __cdecl property(const char* name, unsigned int index, TGCVariant value);
};

//----------------- CConnectionInstance --------------------------------------------------------------------------------

typedef enum tagConnectionPart
{
  GC_CONNECTION_PART_1,
  GC_CONNECTION_PART_CENTER,
  GC_CONNECTION_PART_2
} TConnectionPart;

/** A concrete instance for a connection. */
class CConnectionInstance: public CGraphicElement
{
private:
  // A private listener class for the parts of a connection instance.
  class CConnectionListener: private CGCListener
  {
    friend class CConnectionInstance;
  protected:
    CConnectionInstance* instance;
  public:
    virtual void __cdecl onAction(CGCBase* sender, CGCBase* origin, TActionType actionType) { };
    virtual void __cdecl onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason);
    virtual void __cdecl onDestroy(CGCBase* sender);
    virtual void __cdecl onError(CGCBase* sender, CGCBase* origin, const char* message);
  };

  CConnection* FConnection;
  CFigureInstance* FEnd1;
  CFigureInstance* FEnd2;
  CConnectionListener FListener;
  GLuint FDisplayList;
  TVertex FEndVertex1;
  TVertex FEndVertex2;
  TConnectionDirection FDirection1;    // Direction of the connection with regard to end point 1.
  TConnectionDirection FDirection2;    // Direction of the connection with regard to end point 2.
  TConnectionLineStyle FLineStyle;     // Style of the connection line.
  GLfloat FColor[4];
  GLfloat FHotColor[4];

  // Decorations and labels.
  CFigureInstance* FEnd1Decoration;
  TVertex FLabel1Position;
  CFigureInstance* FEnd1Label;
  CFigureInstance* FCenterDecoration;
  TVertex FCenterLabelPosition;
  CFigureInstance* FCenterLabel;
  CFigureInstance* FEnd2Decoration;
  TVertex FLabel2Position;
  CFigureInstance* FEnd2Label;
  float FCenter;                       // A percentual value reaching from 0 through 1 specifying the entire range 
                                        // from start to end point.
protected:
  void applyTransformations(CFigureInstance* target, TVertex* position, TConnectionDirection direction);
  float computeCenterPosition(void);
  TConnectionDirection determineCenterDirection(void);
  void renderDecoration(TConnectionPart part, float currentZoom);
  void updateDecoration(TConnectionPart part);
  void updateFigureInstance(CFigureInstance** instance, CFigure* figure);
public:
  CConnectionInstance(CConnection* connection, CFigureInstance* endPoint1, CFigureInstance* endPoint2);
  virtual ~CConnectionInstance(void);

  virtual bool __cdecl containsPoint(const float x, const float y);
  virtual float __cdecl center(void) { return FCenter; };
  void computeCoordinates(CFigureInstance* point, TConnectionDirection direction, float slot);
  virtual CFigureInstance* __cdecl endPoint1(void) { return FEnd1; };
  virtual CFigureInstance* __cdecl endPoint2(void) { return FEnd2; };
  TConnectionDirection getDirection(CFigureInstance* point);
  virtual TFeedbackInfo __cdecl getFeedbackInfo(const float x, const float y);
  virtual void __cdecl lineStyle(TConnectionLineStyle newStyle);
  virtual TConnectionLineStyle __cdecl lineStyle(void) { return FLineStyle; };
  void moveCenterTo(TVertex position);
  virtual TGCVariant __cdecl property(const char* name, unsigned int index);
  virtual void __cdecl property(const char* name, unsigned int index, TGCVariant value);
  void render(float currentZoom, bool isHot);
  void updateDecorations(void);
  virtual void __cdecl validate(void);
};

//----------------------------------------------------------------------------------------------------------------------

#endif // __GC_CONNECTION_H__
