/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_layer.h 
 * @brief Implementation of the GC layer class.
 * 
 */

#ifndef __GC_LAYER_H__
#define __GC_LAYER_H__

#include "myx_gc_figure.h"
#include "myx_gc_gl_helper.h"
#include "myx_gc_connection.h"

//----------------------------------------------------------------------------------------------------------------------

class CGenericCanvas;
class CFigureInstance;
class CFigure;
class CHitResults;
class CConnection;
class CConnectionInstance;

/**
 * This is the base layer class, which is used by views to display their content. There are descendants for special things
 * like feedback, grids and so on.
 */
class GENERIC_CANVAS_API CLayer: public CGCBase
{
  friend class CFigureInstance;
  friend class CFigureInstanceEnumerator;
  friend class CInstanceListener;
private:
  class CInstanceListener: private CGCListener
  {
    friend class CLayer;
  protected:
    CLayer* layer;
  public:
    virtual void __cdecl onAction(CGCBase* sender, CGCBase* origin, TActionType action);
    virtual void __cdecl onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason);
    virtual void __cdecl onDestroy(CGCBase* object);
    virtual void __cdecl onError(CGCBase* sender, CGCBase* origin, const char* message);
  };

  wstring FName;                  // The name of the layer.
  bool FDirty;                    // True if any of the properties changed that affect the display list.
  float FScaling[3];              // The factors to scale the layer in each direction.
  float FMinimalScale;            // The lesser of the x and y scale factors.
  float FTranslation[3];          // The factors to move the layer.
  CFigureInstances FInstances;    // A list of figures instances.
  CFigureInstances FGroups;       // A list of groups.
  bool FVisible;                  // true if the layer is visible, otherwise false. Default is true.
  bool FEnabled;                  // True if the layer answers to mouse and keyboard actions.
  CInstanceListener FListener;
protected:
  void applyTransformations();
  void checkError(void);
  void makeDirty(void);
  void renderFeedback(CFigureInstance* instance);
  virtual void renderLayerContent(float currentZoom, TBoundingBox bounds);
  void validate(void);
  virtual void validateLayerContent(void);
public:
  CLayer(string name, CGenericCanvas* canvas);
  virtual ~CLayer(void);

  virtual void __cdecl addInstance(CFigureInstance* instance);
  virtual void __cdecl bringToFront(CFigureInstance* instance, CFigureInstance* relativeTo = NULL);
  virtual void __cdecl clear();
  virtual CFigureInstance* __cdecl createInstance(CFigure* figure);
  void getHitTestInfoAt(CHitResults* hits, TVertex point, bool singleHit);
  virtual void __cdecl enabled(bool isEnabled);
  virtual bool __cdecl enabled(void);
  virtual TVertex __cdecl layerToView(TVertex coords);
  wstring name(void) { return FName; };
  virtual TGCVariant __cdecl property(const char* name, unsigned int index);
  virtual void __cdecl property(const char* name, unsigned int index, TGCVariant value);
  virtual void __cdecl removeInstance(CFigureInstance* instance);
  virtual void __cdecl render(float currentZoom, TBoundingBox bounds);
  virtual void __cdecl scale(const float Factor[3], bool accumulative = false);
  virtual void __cdecl scale(float Sx, float Sy, float Sz, bool accumulative = false);
  virtual void __cdecl sendToBack(CFigureInstance* instance, CFigureInstance* relativeTo = NULL);
  virtual void __cdecl translate(float Tx, float Ty, float Tz, bool accumulative = false);
  virtual void __cdecl translateV(const float Factor[3], bool accumulative = false);
  virtual TVertex __cdecl viewToLayer(TVertex coords);
  virtual void __cdecl visible(bool isVisible);
  virtual bool __cdecl visible(void);
};

//----------------------------------------------------------------------------------------------------------------------

/**
 * The grid layer is a special layer variant that renders itself as grid.
 */
class GENERIC_CANVAS_API CGridLayer: public CLayer
{
private:
  class CViewListener: public CGCListener
  {
    friend class CGridLayer;
  protected:
    CGridLayer* layer;
  public:
    virtual void __cdecl onAction(CGCBase* sender, CGCBase* origin, TActionType action) { };
    virtual void __cdecl onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason);
    virtual void __cdecl onDestroy(CGCBase* object);
    virtual void __cdecl onError(CGCBase* sender, CGCBase* origin, const char* message);
  };

  CGCView* FView;       // The view to which this layer belongs.
  CViewListener FListener;
protected:
  virtual void renderLayerContent(float currentZoom, TBoundingBox bounds);
public:
  CGridLayer(CGCView* view);
  virtual ~CGridLayer(void);

};

//----------------- Selection layer and associated structures ----------------------------------------------------------

typedef struct tagSelectionEntry
{
  CFigureInstance* instance;
  bool dirty;
  TBoundingBox bounds;
} TSelectionEntry;

typedef map<CFigureInstance*, TSelectionEntry*> CSelection;
typedef map<CFigureInstance*, TSelectionEntry*>::iterator CSelectionIterator;
typedef map<CFigureInstance*, TSelectionEntry*>::reverse_iterator CSelectionIteratorReverse;

/** Interal states of the selection layer. */
#define GC_FBSTATE_RUBBERRECT 0x0001
#define GC_FBSTATE_RUBBERBAND 0x0002

class CFeedbackLayer;

/**
 * Helper class to allow enumeration of the currently selected figure instances. An instance is created by the feedback layer's
 * getSelectionEnumerator method and must be freed by the caller. Non C++ callers must use the enumerator's release method to free
 * the class.
 */
class GENERIC_CANVAS_API CSelectionEnumerator
{
private:
  CFeedbackLayer* FLayer;
  CSelectionIterator FIterator;
public:
  CSelectionEnumerator(CFeedbackLayer* layer);

  virtual bool __cdecl hasNext(void);
  virtual CFigureInstance* __cdecl next(void);
  virtual void __cdecl release(void);
  virtual void __cdecl reset(void);
};

/**
 * The selection layer is a special layer variant that renders decorations for selected figures and can be queried
 * for quick hit tests and lists of selected figures.
 */
class GENERIC_CANVAS_API CFeedbackLayer: public CLayer
{
  friend class CSelectionEnumerator;
private:
  class CFeedbackInstanceListener: public CGCListener
  {
    friend class CFeedbackLayer;
  protected:
    CFeedbackLayer* layer;
  public:
    virtual void __cdecl onAction(CGCBase* sender, CGCBase* origin, TActionType action) { };
    virtual void __cdecl onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason);
    virtual void __cdecl onDestroy(CGCBase* object);
    virtual void __cdecl onError(CGCBase* sender, CGCBase* origin, const char* message);
  };

  CGCView* FView;                 // The view to which this layer belongs.
  CSelection FSelection;          // The list of currently selected figure instances.
  float FHandleSize;              // The size of the decoration handles.
  unsigned int FStates;          // A number of flags holding transient states of the layer.
  TBoundingBox FRubberRectBounds; // The current bounds of the rubber band (if active).
  GLuint FRubberRectDisplayList;  // The display list for the rubber band (if active).
  GLuint FSelectionDecoration;    // The display list for the selection decoration.
  CFeedbackInstanceListener FListener;
protected:
  void createSelectionDecoration(void);
  bool internalAddToSelection(CFigureInstance* instance);
  void internalRemoveFromSelection(CFigureInstance* instance);
  virtual void renderLayerContent(float currentZoom, TBoundingBox bounds);
  virtual void validateLayerContent(void);
public:
  CFeedbackLayer(CGCView* view);
  virtual ~CFeedbackLayer(void);

  void addToSelection(CFigureInstance* instance);
  void clearSelection(void);
  virtual TFeedbackInfo getFeedbackInfo(float localX, float localY, float zoom, CFigureInstance** instance);
  CSelectionEnumerator* getSelectionEnumerator(void);
  void invalidateBounds(CFigureInstance* instance);
  void moveSelectedInstances(float x, float y, float z, bool accumulative);
  virtual void __cdecl removeInstance(CFigureInstance* instance);
  void removeFromSelection(CFigureInstance* instance);
  void rubberBandResize(const TVertex& coords, TRRSelectionAction action);
  void rubberBandStart(TRubberBandStyle style, const TVertex& coords);
  void rubberBandStop(void);
  TBoundingBox rubberBounds(void) { return FRubberRectBounds; };
  void rubberRectResize(const TVertex& coords, TRRSelectionAction action);
  void rubberRectStart(TRubberRectStyle style, const TVertex& coords, bool removeSelection);
  void rubberRectStop(void);
};

//----------------- Connection layer and associated structures ---------------------------------------------------------

typedef set<CConnectionInstance*> CConnectionInstanceList;
typedef hash_map<CFigureInstance*, CConnectionInstanceList> CFigureConnectionList;

#define GC_CLAYER_STATE_CENTER_DRAGGING 0x0001   // A center line of a connection is being dragged.
#define GC_CLAYER_STATE_VDRAGGING       0x0002   // A vertical center line of a connection is being dragged.
#define GC_CLAYER_STATE_HDRAG_PENDING   0x0004   // A horizontal center line dragging is about to start.
#define GC_CLAYER_STATE_VDRAG_PENDING   0x0008   // A vertical center line dragging is about to start.

/**
 * The connection layer is a special layer variant that renders connections between figures.
 */
class GENERIC_CANVAS_API CConnectionLayer: public CLayer
{
private:
  // A private listener class.
  class CConnectionInstanceListener: public CGCListener
  {
    friend class CConnectionLayer;
  protected:
    CConnectionLayer* layer;
  public:
    virtual void __cdecl onAction(CGCBase* sender, CGCBase* origin, TActionType action) { };
    virtual void __cdecl onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason);
    virtual void __cdecl onDestroy(CGCBase* object);
    virtual void __cdecl onError(CGCBase* sender, CGCBase* origin, const char* message);
  };

  unsigned int FStates;                         // Contains active states of the layer.
  CConnectionInstanceList FConnections;          // All defined connections.
  CFigureConnectionList FConnectionPointList;    // A list of figure instances with their associated connection instances.
  CConnectionInstanceListener FListener;
  CConnectionInstance* FHotConnection;           // When hot tracking is active then this member contains the connection under the mouse pointer.
  CGCView* FView;                                // The view to which this layer belongs.
  CElementLookupCache* FCache;                   // Used for quick hit info determination.

  // Info related to mouse input.
  CConnectionInstance* FLastHit;                 // The figure instance under the mouse during the last mouse down action.
  int FLastWindowX;                              // Last hit's window coordinates.
  int FLastWindowY;  
  float FLastViewX;                              // Last hit's view coordinates (window coords transformed to view space).
  float FLastViewY;
  bool FDragSelection;                           // If true then the entire selection is dragged, otherwise only the last hit.
  TFeedbackInfo FCurrentFeedback;                 // Used when a mouse action is in progress to avoid validating the display info 
                                                 // frequently.
protected:
  CHitResults* getHitTestInfoAt(TVertex point, bool singleHit);
  void invalidateEndPoint(CFigureInstance* point);
  void validateEndPoint(CFigureInstance* point);
  virtual void renderLayerContent(float currentZoom, TBoundingBox bounds);
  void updateCache(void);
  virtual void validateLayerContent(void);
public:
  CConnectionLayer(CGCView* view);
  virtual ~CConnectionLayer(void);

  CConnectionInstance* createInstance(CConnection* connection, CFigureInstance* endPoint1, CFigureInstance* endPoint2);
  TFeedbackInfo getFeedbackInfo(TVertex position);
  bool handleMouseDown(TMouseButton button, int modifiers, int windowX, int windowY, TVertex& viewCoords);
  bool handleMouseMove(int modifiers, int windowX, int windowY, TVertex& viewCoords);
  bool handleMouseUp(TMouseButton button, int modifiers, int windowX, int windowY, TVertex& viewCoords);
  void invalidateInstances(CConnectionInstance* instance);
  void removeInstance(CConnectionInstance* instance);
  void workspaceChanged(void);
};

//----------------------------------------------------------------------------------------------------------------------

#endif // __GC_LAYER_H__
