/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_connection.cpp 
 * @brief Implementation of the connection class.
 * 
 */

#include "myx_gc_connection.h"

//----------------- CFigureListener ------------------------------------------------------------------------------------

void CConnection::CFigureListener::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
}

//----------------------------------------------------------------------------------------------------------------------

void CConnection::CFigureListener::onDestroy(CGCBase* object)
{
  connection->figureDestroyed((CFigure*) object);
}

//----------------------------------------------------------------------------------------------------------------------

void CConnection::CFigureListener::onError(CGCBase* sender, CGCBase* origin, const char* message)
{
  // Let the model report any errors up the chain. This is not our business.
}

//----------------- CConnection ----------------------------------------------------------------------------------------

CConnection::CConnection(CGenericCanvas* canvas, CFigure* endPoint1, CFigure* endPoint2): CGraphicElement(canvas)
{
  _className = "CConnection";
  FListener.connection = this;
  FEnd1 = endPoint1;
  FEnd1->addListener(&FListener);
  FEnd2 = endPoint2;
  FEnd2->addListener(&FListener);

  FDecorations.end1Decoration = NULL;
  FDecorations.end1Label = NULL;
  FDecorations.end2Decoration = NULL;
  FDecorations.end2Label = NULL;
  FDecorations.centerDecoration = NULL;
  FDecorations.centerLabel = NULL;
}

//----------------------------------------------------------------------------------------------------------------------

CConnection::~CConnection(void)
{
  FEnd1->removeListener(&FListener);
  FEnd2->removeListener(&FListener);

  if (FDecorations.end1Decoration != NULL)
    FDecorations.end1Decoration->removeListener(&FListener);
  if (FDecorations.end1Label != NULL)
    FDecorations.end1Label->removeListener(&FListener);
  if (FDecorations.end2Decoration != NULL)
    FDecorations.end2Decoration->removeListener(&FListener);
  if (FDecorations.end2Label != NULL)
    FDecorations.end2Label->removeListener(&FListener);
  if (FDecorations.centerDecoration != NULL)
    FDecorations.centerDecoration->removeListener(&FListener);
  if (FDecorations.centerLabel != NULL)
    FDecorations.centerLabel->removeListener(&FListener);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets all decoration info in one call.
 *
 * @param data A list of CFigure instances that are to be used as decoration for the connection.
 *             Some or all entries can be NULL indicating so there should be no decoration for the corresponding part.
 */
void CConnection::decorations(const TConnectionDecorations& data)
{
  if (FDecorations.end1Decoration != NULL)
    FDecorations.end1Decoration->removeListener(&FListener);
  if (FDecorations.end1Label != NULL)
    FDecorations.end1Label->removeListener(&FListener);
  if (FDecorations.end2Decoration != NULL)
    FDecorations.end2Decoration->removeListener(&FListener);
  if (FDecorations.end2Label != NULL)
    FDecorations.end2Label->removeListener(&FListener);
  if (FDecorations.centerDecoration != NULL)
    FDecorations.centerDecoration->removeListener(&FListener);
  if (FDecorations.centerLabel != NULL)
    FDecorations.centerLabel->removeListener(&FListener);

  FDecorations = data;

  if (FDecorations.end1Decoration != NULL)
    FDecorations.end1Decoration->addListener(&FListener);
  if (FDecorations.end1Label != NULL)
    FDecorations.end1Label->addListener(&FListener);
  if (FDecorations.end2Decoration != NULL)
    FDecorations.end2Decoration->addListener(&FListener);
  if (FDecorations.end2Label != NULL)
    FDecorations.end2Label->addListener(&FListener);
  if (FDecorations.centerDecoration != NULL)
    FDecorations.centerDecoration->addListener(&FListener);
  if (FDecorations.centerLabel != NULL)
    FDecorations.centerLabel->addListener(&FListener);

  change(this, GC_CHANGE_CONNECTION_DECORATION);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Called by the figure listener notifying the connection about the destruction of a figure it is using.
 *
 * @param figure The figure that is being destroyed.
 */
void CConnection::figureDestroyed(CFigure* figure)
{
  // Deleting one of the endpoints means the connection is no longer valid, hence destroy it in that case.
  if (FEnd1 == figure || FEnd2 == NULL)
    delete this;
  else
    if (FDecorations.end1Decoration == figure)
      FDecorations.end1Decoration = NULL;
    else
      if (FDecorations.end1Label == figure)
        FDecorations.end1Label = NULL;
      else
        if (FDecorations.centerDecoration == figure)
          FDecorations.centerDecoration = NULL;
        else
          if (FDecorations.centerLabel == figure)
            FDecorations.centerLabel = NULL;
          else
            if (FDecorations.end2Decoration == figure)
              FDecorations.end2Decoration = NULL;
            else
              if (FDecorations.end2Label == figure)
                FDecorations.end2Label = NULL;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Retrieves the value of the property given by path. The path syntax is must be something like (here expressed as regex)
 * (container)*(property), where container is a slash and the name of a container class (e.g. layers, figures) and
 * property is the name of a simple property of that container.
 *
 * @param name The name of the property to return.
 * @param index The index of the sub property to return if it is located in a list.
 * @return A description of the property value and, if the property is simple, the actual value.
 */
TGCVariant CConnection::property(const char* name, unsigned int index)
{
  TGCVariant result;

  switch (getContainerID(name))
  {
    case GC_CONTAINER_UNKNOWN:
      {
        switch (getPropertyID(name))
        {
          case GC_PROPERTY_NAME:
            {
              // place holder
              break;
            };
        };

        break;
      };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the value of the given property, which must be a simple property.
 *
 * @param name The name of the property.
 * @param index The index of the sub property to return if it is located in a list.
 * @param value The new value of the property. Automatic conversion is performed where possible.
 */
void CConnection::property(const char* name, unsigned int index, TGCVariant value)
{
  // There are currently no properties that could be changed.
}

//----------------- CConnectionListener --------------------------------------------------------------------------------

void CConnectionInstance::CConnectionListener::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
  switch (reason)
  {
    case GC_CHANGE_CAPTION_PROPERTY:
      break; // Ignore
    case GC_CHANGE_CONNECTION_DECORATION:
      {
        instance->updateDecorations();
        break;
      }
    default:
      instance->makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
  };
}

//----------------------------------------------------------------------------------------------------------------------

void CConnectionInstance::CConnectionListener::onDestroy(CGCBase* object)
{
  delete instance;
}

//----------------------------------------------------------------------------------------------------------------------

void CConnectionInstance::CConnectionListener::onError(CGCBase* sender, CGCBase* origin, const char* message)
{
  // Let the model report any errors up the chain. This is not our business.
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Constructor of the class.
 *
 * @param canvas The canvas to which this class belongs.
 * @param connection The connection which is instantiated in this class.
 * @param endPoint1 The first endpoint.
 * @param endPoint2 The second endpoint.
 */
CConnectionInstance::CConnectionInstance(CConnection* connection, CFigureInstance* endPoint1, CFigureInstance* endPoint2)
  :CGraphicElement(connection->canvas())
{
  _className = "CConnectionInstance";
  FListener.instance = this;
  FConnection = connection;
  FConnection->addListener(&FListener);
  FEnd1 = endPoint1;
  FEnd1->addListener(&FListener);
  FEnd2 = endPoint2;
  FEnd2->addListener(&FListener);
  FDisplayList = 0;
  FDirection1 = GC_DIR_NONE;
  FDirection2 = GC_DIR_NONE;
  FLineStyle = GC_CONNECTION_STYLE_SOLID;
  FColor[0] = 0;
  FColor[1] = 0;
  FColor[2] = 0;
  FColor[3] = 1;
  FHotColor[0] = 0;
  FHotColor[1] = 0;
  FHotColor[2] = 1;
  FHotColor[3] = 1;

  // Decorations and labels.
  FEnd1Decoration = NULL;
  FEnd1Label = NULL;
  FCenterDecoration = NULL;
  FCenterLabel = NULL;
  FEnd2Decoration = NULL;
  FEnd2Label = NULL;
  FCenter = 0.5f;

  updateDecorations();
}

//----------------------------------------------------------------------------------------------------------------------

CConnectionInstance::~CConnectionInstance(void)
{
  FConnection->removeListener(&FListener);
  FEnd1->removeListener(&FListener);
  FEnd2->removeListener(&FListener);

  // Decorations and labels.
  if (FEnd1Decoration != NULL)
  {
    FEnd1Decoration->removeListener(&FListener);
    delete FEnd1Decoration;
  };
  if (FEnd1Label != NULL)
  {
    FEnd1Label->removeListener(&FListener);
    delete FEnd1Label;
  };
  if (FCenterDecoration != NULL)
  { 
    FCenterDecoration->removeListener(&FListener);
    delete FCenterDecoration;
  };
  if (FCenterLabel != NULL)
  { 
    FCenterLabel->removeListener(&FListener);
    delete FCenterLabel;
  };
  if (FEnd2Decoration != NULL)
  {  
    FEnd2Decoration->removeListener(&FListener);
    delete FEnd2Decoration;
  };
  if (FEnd2Label != NULL)
  {
    FEnd2Label->removeListener(&FListener);
    delete FEnd2Label;
  };

  if (FDisplayList != 0)
    glDeleteLists(FDisplayList, 1);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines the correct orientation of the target element and applies all necessary transformations.
 *
 * @param target The figure instance that must be transformed.
 * @param position The position to which the target must be moved.
 * @param direction The orientation of the target. This is used to rotation or mirror the figure instance accordingly.
 */
void CConnectionInstance::applyTransformations(CFigureInstance* target, TVertex* position, TConnectionDirection direction)
{
  target->disableEvents();
  target->translate(position->x, position->y, 0);

  switch (direction)
  {
    case GC_DIR_EAST:
      {
        target->rotate(0, 0, 0, 1);
        target->scale(1, 1, 1);
        break;
      };
    case GC_DIR_SOUTH:
      {
        target->rotate(90, 0, 0, 1);
        target->scale(1, 1, 1);
        break;
      };
    case GC_DIR_WEST:
      {
        target->rotate(0, 0, 0, 1);
        target->scale(-1, 1, 1);
        break;
      };
    case GC_DIR_NORTH:
      {
        target->rotate(-90, 0, 0, 1);
        target->scale(1, 1, 1);
        break;
      };
  };

  target->enableEvents();
  target->validate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Computes the current center coordinate depending on the connection orientation, the relative center value and the
 * current end points.
 */
float CConnectionInstance::computeCenterPosition(void)
{
  float result = 0;

  switch (FDirection1)
  {
    case GC_DIR_SOUTH:
    case GC_DIR_NORTH:
    {
      if (FEndVertex1.y <= FEndVertex2.y)
        result = FEndVertex1.y + FCenter * (FEndVertex2.y - FEndVertex1.y);
      else
        result = FEndVertex2.y + FCenter * (FEndVertex1.y - FEndVertex2.y);

      break;
    }
    default:
    {
      // GC_DIR_EAST/GC_DIR_WEST
      if (FEndVertex1.x <= FEndVertex2.x)
        result = FEndVertex1.x + FCenter * (FEndVertex2.x - FEndVertex1.x);
      else
        result = FEndVertex2.x + FCenter * (FEndVertex1.x - FEndVertex2.x);

      break;
    };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines the visual orienation of the center decoration, which depends not only on the direction of the 
 * connection but also the relation of its end points.
 */
TConnectionDirection CConnectionInstance::determineCenterDirection(void)
{
  TConnectionDirection result = GC_DIR_NONE;
  TBoundingBox bounds = FCenterDecoration->bounds();
  float centerWidth = bounds.lower.x - bounds.upper.x;
  float dX = FEndVertex2.x - FEndVertex1.x;
  float dY = FEndVertex2.y - FEndVertex1.y;

  // For the center we consider the connection to be oriented in a certain direction (from end point 1 to 2).
  switch (FDirection1)
  {
    case GC_DIR_NORTH:
      {
        // If the second end point is to the right of the first and its horizontal distance is large enough
        // then the center points east. If there is not enough distance to make up a meaningful horizontal
        // direction then point the center upwards. Finally if the second point is placed to the left of the
        // first point the center points to west. Similar for all other cases.
        if (fabsf(dX) < centerWidth)
          result = GC_DIR_NORTH;
        else
          if (dX > 0)
            result = GC_DIR_EAST;
          else
            result = GC_DIR_WEST;
        break;
      };
    case GC_DIR_EAST:
      {
        if (fabsf(dY) < centerWidth)
          result = GC_DIR_EAST;
        else
          if (dY > 0)
            result = GC_DIR_SOUTH;
          else
            result = GC_DIR_NORTH;
        break;
      };
    case GC_DIR_SOUTH:
      {
        if (fabsf(dX) < centerWidth)
          result = GC_DIR_SOUTH;
        else
          if (dX > 0)
            result = GC_DIR_EAST;
          else
            result = GC_DIR_WEST;
        break;
      };
    case GC_DIR_WEST:
      {
        if (fabsf(dY) < centerWidth)
          result = GC_DIR_WEST;
        else
          if (dY > 0)
            result = GC_DIR_SOUTH;
          else
            result = GC_DIR_NORTH;
        break;
      };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Renders the given decoration and label if they are assigned.
 *
 * @param part The part to render.
 * @param currentZoom Needed for figure instance rendering.
 */
void CConnectionInstance::renderDecoration(TConnectionPart part, float currentZoom)
{
  CFigureInstance* decoration = NULL;
  CFigureInstance* label = NULL;

  switch (part)
  {
    case GC_CONNECTION_PART_1:
      {
        decoration = FEnd1Decoration;
        label = FEnd1Label;
        break;
      };
    case GC_CONNECTION_PART_CENTER:
      {
        decoration = FCenterDecoration;
        label = FCenterLabel;
        break;
      };
    case GC_CONNECTION_PART_2:
      {
        decoration = FEnd2Decoration;
        label = FEnd2Label;
        break;
      };
  };

  glPushAttrib(GL_ENABLE_BIT);
  glEnable(GL_LINE_SMOOTH);

  if (decoration != NULL)
    decoration->render(currentZoom);
  if (label != NULL)
    label->render(currentZoom);

  glPopAttrib();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Computes the correct position and orientation of the decoration part indicated by part.
 *
 * @param part Determines the part (ends or center) to update.
 */
void CConnectionInstance::updateDecoration(TConnectionPart part)
{
  // niceDistance is a small offset to avoid packing elements too tightly.
  float niceDistance = 2;
  TBoundingBox decorationBounds;

  if (part == GC_CONNECTION_PART_CENTER)
  {
    if (FDirection1 == GC_DIR_NORTH || FDirection1 == GC_DIR_SOUTH)
    {
      if (FCenterDecoration != NULL)
      {
        TVertex position;
        position.x = (FEndVertex1.x + FEndVertex2.x) / 2;
        position.y = computeCenterPosition();
        applyTransformations(FCenterDecoration, &position, determineCenterDirection());
        decorationBounds = FCenterDecoration->bounds();
      };
      if (FCenterLabel != NULL)
      {
        TBoundingBox labelBounds = FCenterLabel->localBounds();
        float labelWidth = labelBounds.lower.x - labelBounds.upper.x;
        float labelHeight = labelBounds.lower.y - labelBounds.upper.y;
        float centerX = (FEndVertex1.x + FEndVertex2.x) / 2 - labelWidth / 2;
        float centerY = decorationBounds.upper.y - labelHeight - niceDistance;
        FCenterLabel->disableEvents();
        FCenterLabel->translate(centerX, centerY, 0);
        FCenterLabel->enableEvents();
      };
    }
    else
    {
      if (FCenterDecoration != NULL)
      {
        TVertex position;
        position.x = computeCenterPosition();
        position.y = (FEndVertex1.y + FEndVertex2.y) / 2;
        applyTransformations(FCenterDecoration, &position, determineCenterDirection());
        decorationBounds = FCenterDecoration->bounds();
      };
      if (FCenterLabel != NULL)
      {
        TBoundingBox labelBounds = FCenterLabel->localBounds();
        float labelHeight = labelBounds.lower.y - labelBounds.upper.y;
        float centerX = decorationBounds.lower.x + niceDistance;
        float centerY = (FEndVertex1.y + FEndVertex2.y) / 2 - labelHeight / 2;
        FCenterLabel->disableEvents();
        FCenterLabel->translate(centerX, centerY, 0);
        FCenterLabel->enableEvents();
      };
    };
  }
  else
  {
    CFigureInstance* decoration;
    CFigureInstance* label;
    TConnectionDirection direction;
    TVertex* position;

    if (part == GC_CONNECTION_PART_1)
    {
      decoration = FEnd1Decoration;
      label = FEnd1Label;
      direction = FDirection1;
      position = &FEndVertex1;
    }
    else
    {
      decoration = FEnd2Decoration;
      label = FEnd2Label;
      direction = FDirection2;
      position = &FEndVertex2;
    };

    // Start with the decoration figure.
    if (decoration != NULL)
      applyTransformations(decoration, position, direction);

    // Next step is the label.
    if (label != NULL)
    {
      label->validate();
      if (decoration != NULL)
        decorationBounds = decoration->bounds();
      TBoundingBox labelBounds = label->localBounds();
      float labelWidth = labelBounds.lower.x - labelBounds.upper.x;
      float labelHeight = labelBounds.lower.y - labelBounds.upper.y;

      float labelX = 0;
      float labelY = 0;

      // Position computation is based on the idea to keep the label as close as possible to the local origin,
      // that is, the coordinate of the end point.
      switch (direction)
      {
        case GC_DIR_NORTH:
          {
            float dX1 = fabsf(position->x - decorationBounds.upper.x);
            float dX2 = fabsf(decorationBounds.lower.x - position->x);
            float dY = fabsf(position->y - decorationBounds.upper.y);
            float x = (dX1 <= dX2) ? decorationBounds.upper.x : decorationBounds.lower.x;
            if (dX1 < dY || dX2 < dY)
            {
              // Left or right to the decoration is closer.
              if (dX1 <= dX2)
                labelX = x -labelWidth - niceDistance; 
              else
                labelX = x + niceDistance; 
              labelY = position->y - labelHeight - niceDistance;
            }
            else
            {
              // Above the decoration is closer.
              labelX = position->x + niceDistance;
              labelY = decorationBounds.lower.y - labelHeight - niceDistance;
            };
            break;
          };
        case GC_DIR_EAST:
          {
            float dY1 = fabsf(decorationBounds.upper.y - position->y);
            float dY2 = fabsf(position->y - decorationBounds.lower.y);
            float dX = fabsf(decorationBounds.lower.x - position->x);
            float y = (dY1 <= dY2) ? decorationBounds.upper.y : decorationBounds.lower.y;
            if (dY1 < dX || dY2 < dX)
            {
              // Above or below the decoration is closer.
              labelX = position->x + niceDistance;
              if (dY1 <= dY2)
                labelY = y - labelHeight - niceDistance; 
              else
                labelY = y + niceDistance; 
            }
            else
            {
              // To the right of the decoration is closer.
              labelX = decorationBounds.lower.x + niceDistance;
              labelY = position->y + niceDistance;
            };
            break;
          };
        case GC_DIR_SOUTH:
          {
            float dX1 = fabsf(position->x - decorationBounds.upper.x);
            float dX2 = fabsf(decorationBounds.lower.x - position->x);
            float dY = fabsf(decorationBounds.lower.y - position->y);
            float x = (dX1 <= dX2) ? decorationBounds.upper.x : decorationBounds.lower.x;
            if (dX1 < dY || dX2 < dY)
            {
              // Left or right to the decoration is closer.
              if (dX1 <= dX2)
                labelX = x - labelWidth - niceDistance; 
              else
                labelX = x + niceDistance; 
              labelY = position->y + niceDistance;
            }
            else
            {
              // Below the decoration is closer.
              labelX = position->x + niceDistance;
              labelY = decorationBounds.lower.y + niceDistance;
            };
            break;
          };
        case GC_DIR_WEST:
          {
            float dY1 = fabsf(decorationBounds.upper.y - position->y);
            float dY2 = fabsf(position->y - decorationBounds.lower.y);
            float dX = fabsf(position->x - decorationBounds.upper.x);
            float y = (dY1 <= dY2) ? decorationBounds.upper.y : decorationBounds.lower.y;
            if (dY1 < dX || dY2 < dX)
            {
              // Above or below the decoration is closer.
              labelX = position->x - labelWidth - niceDistance;
              if (dY1 <= dY2)
                labelY = y - labelHeight - niceDistance; 
              else
                labelY = y + niceDistance; 
            }
            else
            {
              // To the left of the decoration is closer.
              labelX = decorationBounds.upper.x - labelWidth - niceDistance;
              labelY = position->y + niceDistance;
            };
            break;
          };
      };
      label->disableEvents();
      label->translate(labelX, labelY, 0);
      label->enableEvents();
    };
  };
};

//----------------------------------------------------------------------------------------------------------------------

/**
 * Helper method for CConnectionInstance::updateDecorations. It checks if the given figure instance still corresponds
 * to the figure and if not updates the instance.
 */
void CConnectionInstance::updateFigureInstance(CFigureInstance** instance, CFigure* figure)
{
 if ((*instance != NULL) && (figure != NULL))
  {
    // Both references exist. Check if they refer to the same figure.
    if ((*instance)->figure() != figure)
    {
      // Our connection has a new figure as decoration/label. Delete the instance we created for it.
      // A new figure instance will be created by the following code then.
      (*instance)->removeListener(&FListener);
      delete *instance;
      *instance = NULL;
    };
  };

  if ((*instance == NULL) != (figure == NULL))
  {
    // One of both entries does not yet/no longer exist.
    if (*instance != NULL)
    {
      (*instance)->removeListener(&FListener);
      delete *instance;
      *instance = NULL;
    }
    else
    {
      // A new decoration/label was added, create a figure instance for that.
      *instance = new CFigureInstance(NULL, figure);
      (*instance)->addListener(&FListener);
    };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Overrides the hit computation of the base class to precise the result further.
 *
 * @param x Horizontal hit coordinate.
 * @param y Vertical hit coordinate.
 */
bool CConnectionInstance::containsPoint(const float x, const float y)
{
  bool result = CGraphicElement::containsPoint(x, y);
  if (result)
    result = getFeedbackInfo(x, y) != GC_FI_NONE;

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Computes the final coordinates of this connection at the given end point (which must be one of the end points
 * of this connection instance).
 *
 * @param point One of the end points of this connection.
 * @param direction Indicates the edge where the connection leaves the end point.
 * @param slot A distribution offset in the range (0..1) that controls the relative position along the edge.
 */
void CConnectionInstance::computeCoordinates(CFigureInstance* point, TConnectionDirection direction, float slot)
{
  TVertex coordinates;

  if (point == FEnd1)
    FDirection1 = direction;
  else
    if (point == FEnd2)
      FDirection2 = direction;

  TBoundingBox bounds = point->bounds();
  switch (direction)
  {
    case GC_DIR_NORTH:
      {
        coordinates.x = (bounds.lower.x - bounds.upper.x) * slot + bounds.upper.x;
        coordinates.y = bounds.upper.y;
        break;
      };
    case GC_DIR_EAST:
      {
        coordinates.x = bounds.lower.x;
        coordinates.y = (bounds.lower.y - bounds.upper.y) * slot + bounds.upper.y;
        break;
      };
    case GC_DIR_SOUTH:
      {
        coordinates.x = (bounds.lower.x - bounds.upper.x) * slot + bounds.upper.x;
        coordinates.y = bounds.lower.y;
        break;
      };
    case GC_DIR_WEST:
      {
        coordinates.x = bounds.upper.x;
        coordinates.y = (bounds.lower.y - bounds.upper.y) * slot + bounds.upper.y;
        break;
      };
  };
  if (point == FEnd1)
    FEndVertex1 = coordinates;
  else
    FEndVertex2 = coordinates;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines in which direction this connection leaves the given point. This depends entirely on the positions of the
 * endpoints.
 *
 * @param point The point to consider, must be one of the end points of this connection.
 * @return One of the direction flags.
 */
TConnectionDirection CConnectionInstance::getDirection(CFigureInstance* point)
{
  TConnectionDirection result = GC_DIR_NONE;

  CFigureInstance* from = NULL;
  CFigureInstance* to = NULL;
  if (point == FEnd1)
  {
    from = FEnd1;
    to = FEnd2;
  }
  else
    if (point == FEnd2)
    {
      from = FEnd2;
      to = FEnd1;
    };

  if (from != NULL && to != NULL)
  {
    TBoundingBox fromBounds = from->bounds();
    TBoundingBox toBounds = to->bounds();

    result = GC_DIR_NORTH;
    float distance = fromBounds.upper.y - toBounds.lower.y;
    if (toBounds.upper.x - fromBounds.lower.x > distance)
    {
      result = GC_DIR_EAST;
      distance = toBounds.upper.x - fromBounds.lower.x;
    };
    if (toBounds.upper.y - fromBounds.lower.y > distance)
    {
      result = GC_DIR_SOUTH;
      distance = toBounds.upper.y - fromBounds.lower.y;
    };
    if (fromBounds.upper.x - toBounds.lower.x > distance)
    {
      result = GC_DIR_WEST;
      distance = toBounds.lower.x - fromBounds.upper.x;
    };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns info about the connection at the specified point.
 *
 * @param x The x coordinate to check against.
 * @param y The y coordinate to check against.
 * @return The requested info.
 */
TFeedbackInfo CConnectionInstance::getFeedbackInfo(const float x, const float y)
{
  TFeedbackInfo result = GC_FI_NONE;

  float distance = 12; // Distance to the line segments to be considered as "on line".
  float centerPoint = computeCenterPosition();

  if (FDirection1 == GC_DIR_NORTH || FDirection1 == GC_DIR_SOUTH)
  {
    if (fabsf(centerPoint - y) <= distance)
      // Close to the center line. No need to check the other dimension. It is implicitly given by the bounding box.
      result = GC_FI_HLINE_MOVE;
    else
    {
      // Check the other two segments.
      if (y < centerPoint)
      {
        // Left to the center.
        float actualX = (FEndVertex1.y <= centerPoint) ? FEndVertex1.x : FEndVertex2.x;
        if (fabsf(actualX - x) <= distance)
          result = GC_FI_CONNECTION;
      }
      else
      {
        // Right to the center
        float actualX = (FEndVertex1.y > centerPoint) ? FEndVertex1.x : FEndVertex2.x;
        if (fabsf(actualX - x) <= distance)
          result = GC_FI_CONNECTION;
      };
    };
  }
  else
  {
    if (fabsf(centerPoint - x) <= distance)
      // Close to the center line. No need to check the other dimension. It is implicitly given by the bounding box.
      result = GC_FI_VLINE_MOVE;
    else
    {
      // Check the other two segments.
      if (x < centerPoint)
      {
        // Left to the center.
        float actualY = (FEndVertex1.x <= centerPoint) ? FEndVertex1.y : FEndVertex2.y;
        if (fabsf(actualY - y) <= distance)
          result = GC_FI_CONNECTION;
      }
      else
      {
        // Right to the center
        float actualY = (FEndVertex1.x > centerPoint) ? FEndVertex1.y : FEndVertex2.y;
        if (fabsf(actualY - y) <= distance)
          result = GC_FI_CONNECTION;
      };
    };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Changes the visual style of the connection lines.
 *
 * @param lineStyle The new line style to apply.
 */
void CConnectionInstance::lineStyle(TConnectionLineStyle newStyle)
{
  if (FLineStyle != newStyle)
  {
    FLineStyle = newStyle;
    makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Moves the current center point to the given position. This position is converted into a procentual value for the
 * center and only one dimension (depending on the orientation of the connection) is considered.
 *
 * @param position The new position for the center.
 */
void CConnectionInstance::moveCenterTo(TVertex position)
{
  switch (FDirection1)
  {
    case GC_DIR_SOUTH:
    case GC_DIR_NORTH:
    {
      if (FEndVertex1.y <= FEndVertex2.y)
        FCenter = (position.y - FEndVertex1.y) / (FEndVertex2.y - FEndVertex1.y);
      else
        FCenter = (position.y - FEndVertex2.y) / (FEndVertex1.y - FEndVertex2.y);

      if (FCenter < 0)
        FCenter = 0;
      if (FCenter > 1)
        FCenter = 1;

      break;
    }
    default:
    {
      // GC_DIR_EAST/GC_DIR_WEST
      if (FEndVertex1.x <= FEndVertex2.x)
        FCenter = (position.x - FEndVertex1.x) / (FEndVertex2.x - FEndVertex1.x);
      else
        FCenter = (position.x - FEndVertex2.x) / (FEndVertex1.x - FEndVertex2.x);

      if (FCenter < 0)
        FCenter = 0;
      if (FCenter > 1)
        FCenter = 1;

      break;
    };
  };
  makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Retrieves the value of the property given by path. The path syntax is must be something like (here expressed as regex)
 * (container)*(property), where container is a slash and the name of a container class (e.g. layers, figures) and
 * property is the name of a simple property of that container.
 *
 * @param name The name of the property to return.
 * @param index The index of the sub property to return if it is located in a list.
 * @return A description of the property value and, if the property is simple, the actual value.
 */
TGCVariant CConnectionInstance::property(const char* name, unsigned int index)
{
  TGCVariant result;

  switch (getContainerID(name))
  {
    case GC_CONTAINER_UNKNOWN:
      {
        switch (getPropertyID(name))
        {
          case GC_PROPERTY_CENTER:
            {
              result = FCenter;
              break;
            };
        };

        break;
      };
    case GC_CONTAINER_COLOR:
      {
        if (index < 4)
        {
          result = FColor[index];
        };
        break;
      };
    case GC_CONTAINER_HOT_COLOR:
      {
        if (index < 4)
        {
          result = FHotColor[index];
        };
        break;
      };
    case GC_CONTAINER_CONNECTION:
      {
        result = FConnection;
        break;
      };
    case GC_CONTAINER_END1:
      {
        result = FEnd1;
        break;
      };
    case GC_CONTAINER_END1_DECORATION:
      {
        result = FEnd1Decoration;
        break;
      };
    case GC_CONTAINER_END1_LABEL:
      {
        result = FEnd1Label;
        break;
      };
    case GC_CONTAINER_END2:
      {
        result = FEnd2;
        break;
      };
    case GC_CONTAINER_END2_DECORATION:
      {
        result = FEnd2Decoration;
        break;
      };
    case GC_CONTAINER_END2_LABEL:
      {
        result = FEnd2Label;
        break;
      };
    case GC_CONTAINER_CENTER_DECORATION:
      {
        result = FCenterDecoration;
        break;
      };
    case GC_CONTAINER_CENTER_LABEL:
      {
        result = FCenterLabel;
        break;
      };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the value of the given property, which must be a simple property.
 *
 * @param name The name of the property.
 * @param index The index of the sub property to return if it is located in a list.
 * @param value The new value of the property. Automatic conversion is performed where possible.
 */
void CConnectionInstance::property(const char* name, unsigned int index, TGCVariant value)
{
  switch (getContainerID(name))
  {
    case GC_CONTAINER_UNKNOWN:
      {
        switch (getPropertyID(name))
        {
          case GC_PROPERTY_CENTER:
            {
              FCenter = value;
              makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
              break;
            };
        };

        break;
      };
    case GC_CONTAINER_COLOR:
      {
        if (index < 4)
        {
          FColor[index] = value;
          makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        };
        break;
      };
    case GC_CONTAINER_HOT_COLOR:
      {
        if (index < 4)
        {
          FHotColor[index] = value;
          makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        };
        break;
      };
    case GC_CONTAINER_CONNECTION:
      {
        FConnection = value.cast<CConnection>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
    case GC_CONTAINER_END1:
      {
        FEnd1 = value.cast<CFigureInstance>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
    case GC_CONTAINER_END1_DECORATION:
      {
        FEnd1Decoration = value.cast<CFigureInstance>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
    case GC_CONTAINER_END1_LABEL:
      {
        FEnd1Label = value.cast<CFigureInstance>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
    case GC_CONTAINER_END2:
      {
        FEnd2 = value.cast<CFigureInstance>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
    case GC_CONTAINER_END2_DECORATION:
      {
        FEnd2Decoration = value.cast<CFigureInstance>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
    case GC_CONTAINER_END2_LABEL:
      {
        FEnd2Label = value.cast<CFigureInstance>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
    case GC_CONTAINER_CENTER_DECORATION:
      {
        FCenterDecoration = value.cast<CFigureInstance>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
    case GC_CONTAINER_CENTER_LABEL:
      {
        FCenterLabel = value.cast<CFigureInstance>();
        makeDirty(GC_CHANGE_CONNECTION_INSTANCE);
        break;
      };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Renders this instance.
 *
 * @param currentZoom The current zoom factor of the owning layer. Needed for the text.
 */
void CConnectionInstance::render(float currentZoom, bool isHot)
{
  // If all goes right then the connection instance must be validated already at this point.
  if (!dirty())
  {
    if (isHot)
      glColor4fv(FHotColor);
    else
      glColor4fv(FColor);
    glCallList(FDisplayList);

    // Decorations
    renderDecoration(GC_CONNECTION_PART_1, currentZoom);
    renderDecoration(GC_CONNECTION_PART_CENTER, currentZoom);
    renderDecoration(GC_CONNECTION_PART_2, currentZoom);
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Called by the listener when something in the connection's decoration changed or any of used figure instance is being 
 * destroyed. Use this notification to (re)build the figure instances we use for decoration.
 */
void CConnectionInstance::updateDecorations(void)
{
  TConnectionDecorations decorations = FConnection->decorations();

  updateFigureInstance(&FEnd1Decoration, decorations.end1Decoration);
  updateFigureInstance(&FEnd1Label, decorations.end1Label);
  updateFigureInstance(&FCenterDecoration, decorations.centerDecoration);
  updateFigureInstance(&FCenterLabel, decorations.centerLabel);
  updateFigureInstance(&FEnd2Decoration, decorations.end2Decoration);
  updateFigureInstance(&FEnd2Label, decorations.end2Label);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Computes all detail elements for this connection depending on computed start and end point as well as other properties.
 */
void CConnectionInstance::validate(void)
{
  if (dirty())
  {
    CGraphicElement::validate();

    if (FDisplayList == 0)
      FDisplayList = glGenLists(1);

    glNewList(FDisplayList, GL_COMPILE);
      glPushMatrix();
      glPushAttrib(GL_ENABLE_BIT);

      switch (FLineStyle)
      {
        case GC_CONNECTION_STYLE_SOLID:
          break;
        case GC_CONNECTION_STYLE_DOTTED:
          {
            glEnable(GL_LINE_STIPPLE);
            glLineStipple (1, 0x3333);
            break;
          };
      };

      glBegin(GL_LINE_STRIP);
      float centerPoint = computeCenterPosition();
      if (FDirection1 == GC_DIR_NORTH || FDirection1 == GC_DIR_SOUTH)
      {
        // We need to distinguish here only if we have a vertical or horizontal run.
        // So checking only one direction is enough.
        glVertex2f(FEndVertex1.x, FEndVertex1.y);
        glVertex2f(FEndVertex1.x, centerPoint);
        glVertex2f(FEndVertex2.x, centerPoint);
        glVertex2f(FEndVertex2.x, FEndVertex2.y);
      }
      else
      {
        glVertex2f(FEndVertex1.x, FEndVertex1.y);
        glVertex2f(centerPoint, FEndVertex1.y);
        glVertex2f(centerPoint, FEndVertex2.y);
        glVertex2f(FEndVertex2.x, FEndVertex2.y);
      };
      glEnd();
      glPopAttrib();
      glPopMatrix();
    glEndList();

    // Valdate decoration and compute the overall bounding box.
    boundsNew();
    boundsAdd(FEndVertex1);
    boundsAdd(FEndVertex2);

    updateDecoration(GC_CONNECTION_PART_1);
    if (FEnd1Decoration != NULL)
      boundsAdd(FEnd1Decoration->bounds());
    if (FEnd1Label != NULL)
      boundsAdd(FEnd1Label->bounds());

    updateDecoration(GC_CONNECTION_PART_CENTER);
    if (FCenterDecoration != NULL)
      FCenterDecoration->validate();
    if (FCenterLabel != NULL)
      FCenterLabel->validate();

    updateDecoration(GC_CONNECTION_PART_2);
    if (FEnd2Decoration != NULL)
      boundsAdd(FEnd2Decoration->bounds());
    if (FEnd2Label != NULL)
      boundsAdd(FEnd2Label->bounds());

    boundsFinished();
  };
}

//----------------------------------------------------------------------------------------------------------------------

