/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * @file myx_gc_layer.cpp 
 * @brief Implementation of the GC layer class.
 * 
 */

#include "myx_gc_layer.h"

//----------------- CInstanceListener ----------------------------------------------------------------------------------

void CLayer::CInstanceListener::onAction(CGCBase* sender, CGCBase* origin, TActionType actionType)
{
  layer->action(origin, actionType);
}

//----------------------------------------------------------------------------------------------------------------------

void CLayer::CInstanceListener::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
  layer->change(origin, reason);
}

//----------------------------------------------------------------------------------------------------------------------

void CLayer::CInstanceListener::onDestroy(CGCBase* object)
{
  layer->removeInstance((CFigureInstance*) object);
}

//----------------------------------------------------------------------------------------------------------------------

void CLayer::CInstanceListener::onError(CGCBase* sender, CGCBase* origin, const char* message)
{
  layer->error(origin, message);
}

//----------------- CLayer ---------------------------------------------------------------------------------------------

CLayer::CLayer(string name, CGenericCanvas* canvas): CGCBase(canvas)
{
  _className = "CLayer";
  FName = utf8ToUtf16(name);

  // Initialize with useful values.
  FScaling[0] = 1;
  FScaling[1] = 1;
  FScaling[2] = 1;

  FTranslation[0] = 0;
  FTranslation[1] = 0;
  FTranslation[2] = 0;

  FDirty = true;
  FVisible = true;
  FEnabled = true;

  FListener.layer = this;
}

//----------------------------------------------------------------------------------------------------------------------

CLayer::~CLayer(void)
{
  beginUpdate();
  if ((canvas() != NULL) && (!canvas()->updating()))
    canvas()->removeLayer(this);
  clear();
  endUpdate();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Applies the layer's transformations for rendering, feedback etc.
 */
void CLayer::applyTransformations()
{
  glTranslatef(FTranslation[0], FTranslation[1], FTranslation[2]);
  glScalef(FScaling[0], FScaling[1], FScaling[2]);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Triggers the error checking of the canvas.
 */
void CLayer::checkError(void)
{
  canvas()->checkError();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Marks the display list for this layer as invalid, hence it will be recreated next time validate is called.
 * If a list already exists then it is freed.
 */
void CLayer::makeDirty(void)
{
  if (!FDirty)
  {
    FDirty = true;
    if (!destroying() && !updating())
      canvas()->refresh();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/*
 * Helper method to determine the transformed vertices of the given figure instance. The layer applies its own
 * transformations and only renders the figure instance.
 *
 * @param instance The figure instance for which feedback data is requested.
 */
void CLayer::renderFeedback(CFigureInstance* instance)
{
  applyTransformations();
  instance->render(1);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Renders layer content that is not determined by figure instances. This method might be overridden by descendants.
 *
 * @param currentZoom The current zoom state of the layer.
 * @param bounds The area of the layer that only needs to be rendered.
 */
void CLayer::renderLayerContent(float currentZoom, TBoundingBox bounds)
{
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates the display list of this figure (and all child figures) if necessary.
 */
void CLayer::validate(void)
{
  if (FDirty)
  {
    FDirty = false;

    FMinimalScale = FScaling[0] < FScaling[1] ? FScaling[0] : FScaling[1];
    validateLayerContent();

    // Give the instances the opportunity to trigger validation of their associated figures.
    for (CFigureInstances::iterator iterator = FInstances.begin(); iterator != FInstances.end(); ++iterator)
      (*iterator)->validate();

    checkError();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Prepares layer content that is not determined by figure instances. This method might be overridden by descendants.
 */
void CLayer::validateLayerContent(void)
{
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * If the given figure instance is currently on this layer then it is moved to the place immediately above
 * the relativeTo instance. If relativeTo is 0, then it's made the top most instance (they are rendered as stored in the instances array).
 *
 * @param instance The instance to bring to front.
 * @param relativeTo Instance it should be placed relative to (ie, it will be put above this one)
 */
void CLayer::bringToFront(CFigureInstance* instance, CFigureInstance* relativeTo)
{
  CFigureInstances::iterator figIterator = find(FInstances.begin(), FInstances.end(), instance);
  if (figIterator != FInstances.end())
  {
    FInstances.erase(figIterator);
    
    if (relativeTo != NULL)
    {  
      CFigureInstances::iterator relIterator = find(FInstances.begin(), FInstances.end(), instance);
      
      if (relIterator != FInstances.end())
      {
        ++relIterator; // go to the item next to the one we'll go on top of
        
        if (relIterator == FInstances.end())
          FInstances.push_back(instance); // this is the last one
        else
          FInstances.insert(relIterator, instance); // insert before the next one
        canvas()->refresh();
        return;
      }
    }
    FInstances.push_back(instance);
    canvas()->refresh();
  }
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Adds the given figure instance to the end of the instance list. If instance belongs to another layer currently 
 * it is removed from the other's instance list first.
 *
 * @param instance The figure instance to add.
 */
void CLayer::addInstance(CFigureInstance* instance)
{
  if (instance->FLayer != this)
  {
    beginUpdate();
    if (instance->FLayer != NULL)
      instance->FLayer->removeInstance(instance);
    instance->addListener(&FListener);
    FInstances.push_back(instance);
    instance->FLayer = this;
    endUpdate();
    change(instance, GC_CHANGE_LAYER_ADD_INSTANCE);
    makeDirty();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes all figure instances from this layer.
 */
void CLayer::clear(void)
{
  beginUpdate();

  for (CFigureInstances::iterator iterator = FInstances.begin(); iterator != FInstances.end(); ++iterator)
  {
    CFigureInstance* instance = *iterator;
    instance->FLayer = NULL;
    delete instance;
  };
  FInstances.clear();
  endUpdate();

  change(this, GC_CHANGE_LAYER_CLEAR);
  makeDirty();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a new instance for the given figure and adds it to this layer.
 *
 * @param Figure The figure for which the instance is to be created.
 * @return A new figure instance.
 */
CFigureInstance* CLayer::createInstance(CFigure* Figure)
{
  beginUpdate();
  CFigureInstance* instance = new CFigureInstance(this, Figure);
  addInstance(instance);
  endUpdate();

  return instance;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Fills the hit results with all figure instances whose bounds contain the given coordinates.
 *
 * @param Hits [out] The hit collection that is updated.
 * @param point The hit point coordinated given in view space.
 * @param singleHit If true only one hit is returned.
 */
void CLayer::getHitTestInfoAt(CHitResults* Hits, TVertex point, bool singleHit)
{
  // Bounding boxes of figure instances are already in layer space coordinates. We only need to convert the
  // given hit point to our local coordinate system.
  TVertex localPoint = viewToLayer(point);

  // Iterate backwards. Figures rendered later lay on top of earlier rendered figures.
  for (CFigureInstances::reverse_iterator iterator = FInstances.rbegin(); iterator != FInstances.rend(); ++iterator)
  {
    CFigureInstance* instance = *iterator;
    if (instance->containsPoint(localPoint.x, localPoint.y))
    {
      Hits->addHit(instance);
      if (singleHit)
        break;
    };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the layer's enabled state.
 *
 * @param IsEnabled Set it to true if you want the layer to be visible.
 */
void CLayer::enabled(bool IsEnabled)
{
  if (FEnabled != IsEnabled)
  {
    FEnabled = IsEnabled;
    makeDirty();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the current enabled state.
 *
 * @return The current enabled state.
 */
bool CLayer::enabled(void)
{
  return FEnabled;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Converts the given coordinates from layer space to view space.
 *
 * @param coords The coordinates to convert to view space.
 * @return Converted coordinates.
 */
TVertex CLayer::layerToView(TVertex coords)
{
  TVertex vertex;
  vertex.x = coords.x * FScaling[0] + FTranslation[0];
  vertex.y = coords.y * FScaling[1] + FTranslation[1];

  return vertex;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Retrieves the value of the property given by path. The path syntax is must be something like (here expressed as regex)
 * (container)*(property), where container is a slash and the name of a container class (e.g. layers, figures) and
 * property is the name of a simple property of that container.
 *
 * @param name The name of the property to return.
 * @param index If the property is a list then this is the index into that list.
 * @return A description of the property value and, if the property is simple, the actual value.
 */
TGCVariant CLayer::property(const char* name, unsigned int index)
{
  TGCVariant result;

  switch (getContainerID(name))
  {
    case GC_CONTAINER_UNKNOWN:
      {
        switch (getPropertyID(name))
        {
          case GC_PROPERTY_NAME:
            {
              result = utf16ToUtf8(FName);
              break;
            };
          case GC_PROPERTY_DESCRIPTION:
            {
              result = "A collection of figure instances.";
              break;
            };
          case GC_PROPERTY_OWNER:
            {
              result = canvas();
              break;
            };
          case GC_PROPERTY_VISIBLE:
            {
              result = FVisible;
              break;
            };
          case GC_PROPERTY_ENABLED:
            {
              result = FEnabled;
              break;
            };
        };
        break;
      };
    case GC_CONTAINER_FIGURE_INSTANCES:
      {
        if (index < FInstances.size())
          result = FInstances[index];
        break;
      };
    case GC_CONTAINER_GROUPS:
      {
        // TODO: Should actually be the groups.
        if (index < FInstances.size())
          result = FInstances[index];
        break;
      };
    case GC_CONTAINER_SCALING:
      {
        if (index < 3)
          result = FScaling[index];
        break;
      };
    case GC_CONTAINER_TRANSLATION:
      {
        if (index < 3)
          result = FTranslation[index];
        break;
      };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the value of the given property, which must be a simple property.
 *
 * @param name The name of the property.
 * @param index If the property is a list then this is the index into that list.
 * @param Value The new value of the property. Automatic conversion is performed where possible.
 */
void CLayer::property(const char* name, unsigned int index, TGCVariant value)
{
  switch (getContainerID(name))
  {
    case GC_CONTAINER_UNKNOWN:
      {
        switch (getPropertyID(name))
        {
          case GC_PROPERTY_NAME:
            {
              FName = utf8ToUtf16(value);
              change(this, GC_CHANGE_LAYER_PROPERTY);

              break;
            };
          case GC_PROPERTY_VISIBLE:
            {
              FVisible = value;
              change(this, GC_CHANGE_LAYER_PROPERTY);
              
              break;
            };
          case GC_PROPERTY_ENABLED:
            {
              FEnabled = value;
              change(this, GC_CHANGE_LAYER_PROPERTY);
              
              break;
            };
        };
        break;
      };
    case GC_CONTAINER_SCALING:
      {
        if (index < 3)
        {
          FScaling[index] = value;
          change(this, GC_CHANGE_LAYER_PROPERTY);
        };

        break;
      };
    case GC_CONTAINER_TRANSLATION:
      {
        if (index < 3)
        {
          FTranslation[index] = value;
          change(this, GC_CHANGE_LAYER_PROPERTY);
        };

        break;
      };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes the given figure instance from the instance list if it is currently there.
 * No error is raised if the instance does not belong to this layer.
 *
 * @param instance The instance to be removed.
 */
void CLayer::removeInstance(CFigureInstance* instance)
{
  if (!updating() && !destroying())
  {
    beginUpdate();
    for (CFigureInstances::iterator iterator = FInstances.begin(); iterator != FInstances.end(); ++iterator)
      if (*iterator == instance)
      {
        (*iterator)->removeListener(&FListener);
        FInstances.erase(iterator);
        instance->FLayer = NULL;
        change(this, GC_CHANGE_LAYER_REMOVE_INSTANCE);
        makeDirty();
        break;
      };
    endUpdate();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Checks the validity of the figure display list and executes it.
 *
 * @param currentZoom The current zoom factor of the view.
 * @param bounds The area currently visible. No need to render anything outside that area.
 */
void CLayer::render(float currentZoom, TBoundingBox bounds)
{
  if (FVisible && !updating()) 
  {
    if (FDirty)
      validate();

    currentZoom *= FMinimalScale;
    glPushMatrix();
    applyTransformations();

    // Transform viewport coordinates into view space. They are used for occlusion culling.
    TBoundingBox visibleBounds;
    visibleBounds.lower = viewToLayer(bounds.lower);
    visibleBounds.upper = viewToLayer(bounds.upper);
    for (CFigureInstances::iterator iterator = FInstances.begin(); iterator != FInstances.end(); ++iterator)
    {
      CFigureInstance* instance = *iterator;
      if (instance->overlaps(visibleBounds))
        instance->render(currentZoom);
    };

    renderLayerContent(currentZoom, visibleBounds);
    glPopMatrix();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Scales the layer by the amount given in Factor. If Accumulative is true then the new scale factors are multiplied
 * with the existing values. This version of scale uses single float values as parameters.
 *
 * @param Sx scale factor for the x axis.
 * @param Sy scale factor for the y axis.
 * @param Sz scale factor for the z axis.
 * @param Accumulative If true then the given values are added to any existing values otherwiese they are used as given.
 */
void CLayer::scale(float Sx, float Sy, float Sz, bool Accumulative)
{
  if (Accumulative)
  {
    FScaling[0] += Sx;
    FScaling[1] += Sy;
    FScaling[2] += Sz;
  }
  else
  {
    FScaling[0] = Sx;
    FScaling[1] = Sy;
    FScaling[2] = Sz;
  };
  change(this, GC_CHANGE_LAYER_PROPERTY);
  makeDirty();
}

//----------------------------------------------------------------------------------------------------------------------

/** 
 * Scales the layer by the amount given in Factor. If Accumulative is true then the new scale factors are multiplied
 * with the existing values. This version of scale uses an array of values in the parameter list.
 *
 * @param Factor An array of 3 scale values, one for each axis.
 * @param Accumulative If true then the given values are added to any existing values otherwiese they are used as given.
 */
void CLayer::scale(const float Factor[3], bool Accumulative)
{
  if (Accumulative)
  {
    FScaling[0] += Factor[0];
    FScaling[1] += Factor[1];
    FScaling[2] += Factor[2];
  }
  else
  {
    FScaling[0] = Factor[0];
    FScaling[1] = Factor[1];
    FScaling[2] = Factor[2];
  };
  change(this, GC_CHANGE_LAYER_PROPERTY);
  makeDirty();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * If the given figure instance is currently on this layer then it is moved to the first place in the list making it so
 * the bottom most instance (they are rendered as stored in the instances array).
 *
 * @param instance The instance to send to back.
 * @param relativeTo The instance it should be moved relative to.
 */
void CLayer::sendToBack(CFigureInstance* instance, CFigureInstance* relativeTo)
{
  CFigureInstances::iterator figIterator = find(FInstances.begin(), FInstances.end(), instance);
  if (figIterator != FInstances.end())
  {
    FInstances.erase(figIterator);
    if (relativeTo != NULL)
    {  
      CFigureInstances::iterator relIterator = find(FInstances.begin(), FInstances.end(), instance);
      
      if (relIterator != FInstances.end() && relIterator != figIterator)
      {
        FInstances.insert(relIterator, instance);
        canvas()->refresh();
        return;
      }
    }
    FInstances.insert(FInstances.begin(), instance); // insert before 1st element
    canvas()->refresh();
  }
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Moves the layer by the amount given in Tx, Ty and Tz. If Accumulative is true then the new translation factors 
 * are multiplied with the existing values. This version of translate uses an array for the values in the parameter list.
 *
 * @param Tx scale factor for the x axis.
 * @param Ty scale factor for the y axis.
 * @param Tz scale factor for the z axis.
 * @param Accumulative If true then the given values are added to any existing values otherwiese they are used as given.
 */
void CLayer::translate(float Tx, float Ty, float Tz, bool Accumulative)
{
  if (Accumulative)
  {
    FTranslation[0] += Tx;
    FTranslation[1] += Ty;
    FTranslation[2] += Tz;
  }
  else
  {
    FTranslation[0] = Tx;
    FTranslation[1] = Ty;
    FTranslation[2] = Tz;
  };
  change(this, GC_CHANGE_LAYER_PROPERTY);
  makeDirty();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Moves the layer by the amount given in Factor. If Accumulative is true then the new translation factors are multiplied
 * with the existing values. This version of translate uses an array for the values in the parameter list.
 *
 * @param Factor An array of translation values, for each axis one.
 * @param Accumulative If true then the given values are added to any existing values otherwiese they are used as given.
 */
void CLayer::translateV(const float Factor[3], bool Accumulative)
{
  if (Accumulative)
  {
    FTranslation[0] += Factor[0];
    FTranslation[1] += Factor[1];
    FTranslation[2] += Factor[2];
  }
  else
  {
    FTranslation[0] = Factor[0];
    FTranslation[1] = Factor[1];
    FTranslation[2] = Factor[2];
  };
  change(this, GC_CHANGE_LAYER_PROPERTY);
  makeDirty();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Converts the given coordinates from view space to layer space.
 *
 * @param coords The coordinates to convert to layer space.
 * @return Converted coordinates.
 */
TVertex CLayer::viewToLayer(TVertex coords)
{
  TVertex vertex;
  vertex.x = (coords.x - FTranslation[0]) / FScaling[0];
  vertex.y = (coords.y - FTranslation[1]) / FScaling[1];

  return vertex;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Sets the layer's visibility state.
 *
 * @param IsVisible Set it to true if you want the layer to be visible.
 */
void CLayer::visible(bool IsVisible)
{
  if (FVisible != IsVisible)
  {
    FVisible = IsVisible;
    change(this, GC_CHANGE_LAYER_VISIBILITY);
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the visibility state.
 *
 * @return The current visibility state.
 */
bool CLayer::visible(void)
{
  return FVisible;
}

//----------------- CViewListener --------------------------------------------------------------------------------------

void CGridLayer::CViewListener::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
  layer->makeDirty();
}

//----------------------------------------------------------------------------------------------------------------------

void CGridLayer::CViewListener::onDestroy(CGCBase* object)
{
  // This event will never occur because the viewer frees its grid layer before it is destroyed.
}

//----------------------------------------------------------------------------------------------------------------------

void CGridLayer::CViewListener::onError(CGCBase* sender, CGCBase* origin, const char* message)
{
  // Let the owner layer report any errors up the chain. This is not our business.
}

//----------------- CGridLayer -----------------------------------------------------------------------------------------

CGridLayer::CGridLayer(CGCView* view): CLayer("grid", view->canvas())
{
  _className = "CGridLayer";
  FListener.layer = this;
  FView = view;
  FView->addListener(&FListener);
}

//----------------------------------------------------------------------------------------------------------------------

CGridLayer::~CGridLayer(void)
{
  FView->removeListener(&FListener);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Renders the grid depending on the current display part and the zoom state.
 *
 * @param currentZoom The zoom factor which determines the grid resolution.
 * @param bounds The area for the grid.
 */
void CGridLayer::renderLayerContent(float currentZoom, TBoundingBox bounds)
{
#define Level3Distance   4
#define Level2Distance  20
#define Level1Distance 100

  if (canvas()->supportsExtension(GC_OE_MULTISAMPLING))
    glDisable(GL_MULTISAMPLE_ARB);

  bool showLevel2 = currentZoom >= 0.5;
  bool showLevel3 = currentZoom >= 2;

  float originX, originY, originZ;

  FView->getOrigin(&originX, &originY, &originZ);

  glColor4f(0.0f, 0.0f, 0.0f, 0.05f);

  if (showLevel3)
  {
    // Third level lines.
    glEnable (GL_LINE_STIPPLE);
    glLineStipple (1, 0x3333);  //  dotted
    glBegin(GL_LINES);

    // The actual paint start points depend on the origin and the area to draw.
    float left = originX - Level3Distance * ceil((originX - bounds.upper.x) / Level3Distance);
    float right = originX + Level3Distance * ceil((bounds.lower.x - originX) / Level3Distance);
    float top = originY - Level3Distance * ceil((originY - bounds.upper.y) / Level3Distance);
    float bottom = originY + Level3Distance * ceil((bounds.lower.y - originY) / Level3Distance);

    for (float y = top; y < bottom; y += Level3Distance)
    {
      glVertex2f(left, y);
      glVertex2f(right, y);
    };
    glVertex2f(left, bottom);
    glVertex2f(right, bottom);

    for (float x = left; x < right; x += Level3Distance)
    {
      glVertex2f(x, top);
      glVertex2f(x, bottom);
    };
    glVertex2f(right, top);
    glVertex2f(right, bottom);
    glEnd();
    glDisable(GL_LINE_STIPPLE);
  };

  if (showLevel2)
  {
    // Second level lines.
    glBegin(GL_LINES);

    // The actual paint start points depend on the origin and the area to draw.
    float left = originX - Level2Distance * ceil((originX - bounds.upper.x) / Level2Distance);
    float right = originX + Level2Distance * ceil((bounds.lower.x - originX) / Level2Distance);
    float top = originY - Level2Distance * ceil((originY - bounds.upper.y) / Level2Distance);
    float bottom = originY + Level2Distance * ceil((bounds.lower.y - originY) / Level2Distance);

    for (float y = top; y < bottom; y += Level2Distance)
    {
      glVertex2f(left, y);
      glVertex2f(right, y);
    };
    glVertex2f(left, bottom);
    glVertex2f(right, bottom);

    for (float x = left; x < right; x += Level2Distance)
    {
      glVertex2f(x, top);
      glVertex2f(x, bottom);
    };
    glVertex2f(right, top);
    glVertex2f(right, bottom);
    glEnd();
  };

  // Top level lines.
  glBegin(GL_LINES);

  // The actual paint start points depend on the origin and the area to draw.
  float left = originX - Level1Distance * ceil((originX - bounds.upper.x) / Level1Distance);
  float right = originX + Level1Distance * ceil((bounds.lower.x - originX) / Level1Distance);
  float top = originY - Level1Distance * ceil((originY - bounds.upper.y) / Level1Distance);
  float bottom = originY + Level1Distance * ceil((bounds.lower.y - originY) / Level1Distance);

  for (float y = top; y < bottom; y += Level1Distance)
  {
    glVertex2f(left, y);
    glVertex2f(right, y);
  };
  glVertex2f(left, bottom);
  glVertex2f(right, bottom);

  for (float x = left; x < right; x += Level1Distance)
  {
    glVertex2f(x, top);
    glVertex2f(x, bottom);
  };
  glVertex2f(right, top);
  glVertex2f(right, bottom);
  glEnd();

  // Finally draw a cross where the origin is located.
  glBegin(GL_LINES);
    glVertex2f(originX, bounds.upper.y);
    glVertex2f(originX, bounds.lower.y);
    glVertex2f(bounds.upper.x, originY);
    glVertex2f(bounds.lower.x, originY);
  glEnd();

  if (canvas()->supportsExtension(GC_OE_MULTISAMPLING))
    glEnable(GL_MULTISAMPLE_ARB);
}

//----------------- CSelectionEnumerator -------------------------------------------------------------------------------

/**
 * Constructor of the enumerator class.
 *
 * @param layer The feedback layer which contains the selection.
 */
CSelectionEnumerator::CSelectionEnumerator(CFeedbackLayer* layer)
{
  FLayer = layer;
  reset();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines if there is a next figure instance to enumerate.
 *
 * @return True if there is still a figure instance otherwise false.
 */
bool CSelectionEnumerator::hasNext(void)
{
  return FIterator != FLayer->FSelection.end();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Returns the next figure instance in the sequence.
 *
 * @return The next figure instance.
 */
CFigureInstance* CSelectionEnumerator::next(void)
{
  CFigureInstance* result = FIterator->first;

  // Advance to next instance.
  ++FIterator;

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Frees this enumerator instance. Usually called by non-C++ languages as memory is managed by the C++ runtime.
 */
void CSelectionEnumerator::release(void)
{
  delete this;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Resets the enumerator to the first figure instance in the canvas.
 */
void CSelectionEnumerator::reset(void)
{
  FIterator = FLayer->FSelection.begin();
}

//----------------- CFeedbackInstanceListener --------------------------------------------------------------------------

void CFeedbackLayer::CFeedbackInstanceListener::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
  layer->invalidateBounds((CFigureInstance*) sender);
}

//----------------------------------------------------------------------------------------------------------------------

void CFeedbackLayer::CFeedbackInstanceListener::onDestroy(CGCBase* object)
{
  layer->removeFromSelection((CFigureInstance*) object);
}

//----------------------------------------------------------------------------------------------------------------------

void CFeedbackLayer::CFeedbackInstanceListener::onError(CGCBase* sender, CGCBase* origin, const char* message)
{
  // Let the owner layer report any errors up the chain. This is not our business.
}

//----------------- CFeedbackLayer ------------------------------------------------------------------------------------

CFeedbackLayer::CFeedbackLayer(CGCView* view): CLayer("feedback", view->canvas())
{
  _className = "CFeedbackLayer";
  FView = view;
  FHandleSize = 5;
  FRubberRectDisplayList = 0;
  FStates = 0;
  FSelectionDecoration = 0;
  FListener.layer = this;
}

//----------------------------------------------------------------------------------------------------------------------

CFeedbackLayer::~CFeedbackLayer(void)
{
  clearSelection();
  if (FSelectionDecoration != 0)
    glDeleteLists(FSelectionDecoration, 1);
  if (FRubberRectDisplayList != 0)
    glDeleteLists(FRubberRectDisplayList, 1);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates the display list for the selection decoration, which is shared among all selection entries.
 */
void CFeedbackLayer::createSelectionDecoration(void)
{
  GLubyte Color[4];
  if (!colorByName("Highlight", Color))
  {
    // If the system's highlight color could not be found then use a default one.
    Color[0] = 64;
    Color[1] = 64;
    Color[2] = 128;
  };

  FSelectionDecoration = glGenLists(1);
  glNewList(FSelectionDecoration, GL_COMPILE);

  // The interior.
  /*
  Color[3] = 30;
  glColor4ubv(Color);
  glRectd(0, 1, 1, 0);
  */
  glPushAttrib(GL_LINE_BIT);
  glDisable(GL_LINE_SMOOTH);

  // The border.
  Color[3] = 200;
  glColor4ubv(Color);
  glBegin(GL_LINE_LOOP);
    glVertex2d(0, 1);
    glVertex2d(1, 1); 
    glVertex2d(1, 0);
    glVertex2d(0, 0); 
  glEnd();

  // The handles.
  Color[3] = 100;
  glColor4ubv(Color);
  glPointSize(FHandleSize);
  glBegin(GL_POINTS);
    glVertex2d(0, 1);
    glVertex2d(0.5, 1);
    glVertex2d(1, 1);

    glVertex2d(0, 0.5);
    glVertex2d(1, 0.5);

    glVertex2d(0, 0);
    glVertex2d(0.5, 0);
    glVertex2d(1, 0);
  glEnd();

  glPopAttrib();

  glEndList();

  checkError();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Helper method to add a figure instance to the selection list. No change event is triggered.
 *
 * @param instance The instance to add.
 * @return If the instance was added (because it wasn't already there) then true is returned, otherwise false.
 */
bool CFeedbackLayer::internalAddToSelection(CFigureInstance* instance)
{
  bool Result = !instance->FSelected;
  if (Result)
  {
    instance->FSelected = true;
    TSelectionEntry* entry = new TSelectionEntry;
    entry->instance = instance;
    entry->dirty = true;
    FSelection[instance] = entry;

    instance->addListener(&FListener);
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Helper method to remove a figure instance from the selection list. No change event is triggered.
 *
 * @param instance The instance to add.
 */
void CFeedbackLayer::internalRemoveFromSelection(CFigureInstance* instance)
{
  if (instance->FSelected)
  {
    if (!instance->destroying())
    {
      instance->FSelected = false;
      instance->removeListener(&FListener);
    };
    CSelectionIterator iterator = FSelection.find(instance);
    if (iterator != FSelection.end())
    {
      delete iterator->second;
      FSelection.erase(iterator);
    };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Renders the decorations for all figure instances that are currently selected.
 */
void CFeedbackLayer::renderLayerContent(float currentZoom, TBoundingBox bounds)
{
  for (CSelectionIterator iterator = FSelection.begin(); iterator != FSelection.end(); iterator++)
  {
    TSelectionEntry* Entry = iterator->second;
    TBoundingBox localBounds;
    localBounds.lower = viewToLayer(bounds.lower);
    localBounds.upper = viewToLayer(bounds.upper);
    if (boundsIntersect(Entry->bounds, localBounds))
    {
      glPushMatrix();
      glTranslated(Entry->bounds.upper.x, Entry->bounds.upper.y, 1);
      glScaled(Entry->bounds.lower.x - Entry->bounds.upper.x, Entry->bounds.lower.y - Entry->bounds.upper.y, 1);
      glCallList(FSelectionDecoration);
      glPopMatrix();
    };
  };
  
  if ((FStates & GC_FBSTATE_RUBBERRECT) != 0)
  {
    // In order to avoid backface culling if the coordinates of the rubber band do not form a counter-clock-wise
    // face we simply disable face culling for the moment.
    glPushAttrib(GL_ENABLE_BIT);
    glDisable(GL_CULL_FACE);
    glTranslated(FRubberRectBounds.upper.x, FRubberRectBounds.upper.y, 0);
    glScaled(FRubberRectBounds.lower.x - FRubberRectBounds.upper.x, FRubberRectBounds.lower.y - FRubberRectBounds.upper.y, 1);
    glCallList(FRubberRectDisplayList);
    glPopAttrib();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates display lists for all invalid decorations.
 */
void CFeedbackLayer::validateLayerContent(void)
{
  if (FSelectionDecoration == 0)
    createSelectionDecoration();

  for (CSelectionIterator iterator = FSelection.begin(); iterator != FSelection.end(); iterator++)
    if (iterator->second->dirty)
    {
      iterator->second->dirty = false;

      // Update the bounding box of the figure instance.
      TBoundingBox bounds = iterator->second->instance->bounds();

      // Give a small border around the figure instance.
      bounds.upper.x -= 1;
      bounds.upper.y -= 1;
      bounds.lower.x += 1;
      bounds.lower.y += 1;
      iterator->second->bounds = bounds;
    };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Adds the given figure instance to the current selection.
 *
 * @param instance The instance to be added to the selection. If it is already in the set it won't be added again.
 */
void CFeedbackLayer::addToSelection(CFigureInstance* instance)
{
  if (internalAddToSelection(instance))
  {
    change(instance, GC_CHANGE_SELECTION_ADD);
    if (!updating())
      makeDirty();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes all figure instances from the selection set, making it empty.
 */
void CFeedbackLayer::clearSelection(void)
{
  if (FSelection.size() > 0)
  {
    for (CSelectionIterator iterator = FSelection.begin(); iterator != FSelection.end(); iterator++)
    {
      CFigureInstance* instance = iterator->second->instance;
      instance->removeListener(&FListener);
      instance->FSelected = false;
      
      delete iterator->second;
    };

    FSelection.clear();
    change(NULL, GC_CHANGE_SELECTION_CLEAR);
    if (!updating())
      makeDirty();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Determines what feedback action could be executed at the given position. The returned info is usually used to 
 * set an indicator (e.g. the mouse pointer) to a certain state to reflect what is possible at that point.
 *
 * @param x The horizontal target position in layer space.
 * @param y The vertical target position in layer space.
 * @param instance [out] Returns the instance that was hit, if any.
 * @return A flag indicating the possible action state.
 */
TFeedbackInfo CFeedbackLayer::getFeedbackInfo(float x, float y, float zoom, CFigureInstance** instance)
{
  TFeedbackInfo Result = GC_FI_NONE;

  // Now try to find a decoration that is located at the given position.
  // We examine only those that are not dirty. dirty decorations cannot be in view.
  // Iteration happens backwards to find decorations on top earlier than others.
  float distance = 4 / zoom;
  for (CSelectionIteratorReverse iterator = FSelection.rbegin(); iterator != FSelection.rend(); ++iterator)
  {
    if (!iterator->second->dirty)
    {
      TBoundingBox bounds = iterator->second->bounds;
      if ((x >= (bounds.upper.x - distance)) && (x <= (bounds.lower.x + distance)) &&
        (y > (bounds.upper.y - distance)) && (y <= (bounds.lower.y + distance)))
      {                                                
        // Found a decoration. Check if a handle is hit.
        *instance = iterator->first;
        Result = GC_FI_SELECTION_BODY;

        bool leftColumn = ::fabs(bounds.upper.x - x) <= distance;
        bool middleColumn = ::fabs((bounds.upper.x + bounds.lower.x) / 2 - x) <= distance;
        bool rightColumn = ::fabs(bounds.lower.x - x) <= distance;
        bool topRow = ::fabs(bounds.upper.y - y) <= distance;
        bool middleRow = ::fabs((bounds.upper.y + bounds.lower.y) / 2 - y) <= distance;
        bool bottomRow = ::fabs(bounds.lower.y - y) <= distance;

        if (leftColumn)
        {
          if (topRow)
            Result = GC_FI_RESIZE_NORTH_WEST;
          else
            if (middleRow)
              Result = GC_FI_RESIZE_WEST;
            else
              if (bottomRow)
                Result = GC_FI_RESIZE_SOUTH_WEST;
        }
        else
          if (middleColumn)
          {
            if (topRow)
              Result = GC_FI_RESIZE_NORTH;
            else
              if (bottomRow)
                Result = GC_FI_RESIZE_SOUTH;
          }
          else
            if (rightColumn)
            {
              if (topRow)
                Result = GC_FI_RESIZE_NORTH_EAST;
              else
                if (middleRow)
                  Result = GC_FI_RESIZE_EAST;
                else
                  if (bottomRow)
                    Result = GC_FI_RESIZE_SOUTH_EAST;
            };

        break;
      };
    };
  };

  return Result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates and returns a selection entry enumerator.
 *
 * @return A new selection enumerator instance. The caller is responsible for freeing it!
 */
CSelectionEnumerator* CFeedbackLayer::getSelectionEnumerator(void)
{
  return new CSelectionEnumerator(this);
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Invalidates the selection decoration of the given instance (or all instances if instance is NULL) so they are recomputed
 * next time the selection layer draws them.
 *
 * @param instance The figure instance whose bounds need recomputation. If this parameter is NULL then all bounds are invalidated.
 */
void CFeedbackLayer::invalidateBounds(CFigureInstance* instance)
{
  if (instance == NULL)
  {
    for (CSelectionIterator iterator = FSelection.begin(); iterator != FSelection.end(); ++iterator)
      iterator->second->dirty = true;
    makeDirty();
  }
  else
    if (instance->FSelected)
    {
      CSelectionIterator iterator = FSelection.find(instance);
      if (iterator != FSelection.end())
      {
        iterator->second->dirty = true;
        makeDirty();
      };
    };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Translates all currently selected figure instances by the given amount.
 *
 * @param x The horizontal amount for moving.
 * @param y The vertical amount for moving.
 * @param z The depth amount for moving.
 * @param accumulative Relative or absolute move.
 */
void CFeedbackLayer::moveSelectedInstances(float x, float y, float z, bool accumulative)
{
  if (FSelection.size() > 0)
  {
    for (CSelectionIterator iterator = FSelection.begin(); iterator != FSelection.end(); iterator++)
    {
      CFigureInstance* instance = iterator->second->instance;
      instance->translate(x, y, z, accumulative);
    };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes the given figure instance from the current selection.
 *
 * @param instance The instance to be removed. If it isn't actually selected then nothing happens.
 */
void CFeedbackLayer::removeFromSelection(CFigureInstance* instance)
{
  if (instance->FSelected)
  {
    internalRemoveFromSelection(instance);
    change(instance, GC_CHANGE_SELECTION_REMOVE);
    if (!updating())
      makeDirty();
  };
}

//----------------------------------------------------------------------------------------------------------------------

void CFeedbackLayer::rubberBandResize(const TVertex& coords, TRRSelectionAction action)
{
}

//----------------------------------------------------------------------------------------------------------------------

void CFeedbackLayer::rubberBandStart(TRubberBandStyle style, const TVertex& coords)
{
}

//----------------------------------------------------------------------------------------------------------------------

void CFeedbackLayer::rubberBandStop(void)
{
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Method from CLayer overriden to also remove the instance from the current selection if necessary.
 *
 * @param instance The instance to be removed.
 */
void CFeedbackLayer::removeInstance(CFigureInstance* instance)
{
  if (updating())
    internalRemoveFromSelection(instance);
  else
    removeFromSelection(instance);
  CLayer::removeInstance(instance);
}

//----------------------------------------------------------------------------------------------------------------------

/**
  * When in rubber rectangle mode then this function extends the current rubber rectangle from the start point to
  * the given coordinates and handles selection/deselection of figure instances.
  *
  * @param coords The coordinate of the new corner in view space.
  * @param Action Determines if and how figure instance selection is to be handled. See TRRSelectionAction for
  *               a description of the various modes.
 */
void CFeedbackLayer::rubberRectResize(const TVertex& coords, TRRSelectionAction Action)
{
  if ((FStates & GC_FBSTATE_RUBBERRECT) != 0)
  {
    TBoundingBox oldBounds = sortBounds(FRubberRectBounds);
    FRubberRectBounds.lower = viewToLayer(coords);
    TBoundingBox newBounds = sortBounds(FRubberRectBounds);

    bool haveSelectionRemoved = false;
    bool haveSelectionAdded = false;

    if (Action != GC_RRACTION_NONE)
    {
      CFigureInstanceEnumerator* enumerator = FView->getFigureInstanceEnumerator();
      switch (Action)
      {
        case GC_RRACTION_SELECT:
          {
            // Select all figure instances if they intersect. Leave all others alone.
            while (enumerator->hasNext())
            {
              CFigureInstance* instance = enumerator->next();
              TBoundingBox instanceBounds = instance->bounds();
              if (!boundsAreEmpty(instanceBounds))
              {
                if (boundsIntersect(newBounds, instanceBounds))
                {
                  internalAddToSelection(instance);
                  haveSelectionAdded = true;
                };
              };
            };
            if (haveSelectionAdded)
              change(NULL, GC_CHANGE_SELECTION_ADD);

            break;
          };
        case GC_RRACTION_SELECT_REMOVE:
          {
            // Select figure instances, which intersect, unselect all others.
            while (enumerator->hasNext())
            {
              CFigureInstance* instance = enumerator->next();
              TBoundingBox instanceBounds = instance->bounds();
              if (!boundsAreEmpty(instanceBounds))
              {
                if (instance->FSelected)
                {
                  // instance is selected. See if it is still within the rubberband bounds.
                  // Remove it from the selection list if not.
                  if (!boundsIntersect(newBounds, instanceBounds))
                  {
                    internalRemoveFromSelection(instance);
                    haveSelectionRemoved = true;
                  };
                }
                else
                {
                  // instance is not selected. Add it to the current selection if its bounds
                  // fall within the rubberband bounds.
                  if (boundsIntersect(newBounds, instanceBounds))
                  {
                    internalAddToSelection(instance);
                    haveSelectionAdded = true;
                  };
                };
              };
            };

            if (haveSelectionRemoved)
              change(NULL, GC_CHANGE_SELECTION_REMOVE);

            if (haveSelectionAdded)
              change(NULL, GC_CHANGE_SELECTION_ADD);

            break;
          };
        case GC_RRACTION_TOGGLE:
          {
            // Switch figure instance selection state if their bounds intersect either
            // the new rubber band bounds or the old, but not both.
            while (enumerator->hasNext())
            {
              CFigureInstance* instance = enumerator->next();
              TBoundingBox instanceBounds = instance->bounds();

              if (!boundsAreEmpty(instanceBounds))
              {
                if (boundsIntersect(newBounds, instanceBounds) != boundsIntersect(oldBounds, instanceBounds))
                {
                  if (instance->FSelected)
                  {
                    internalRemoveFromSelection(instance);
                    haveSelectionRemoved = true;
                  }
                  else
                  {
                    internalAddToSelection(instance);
                    haveSelectionAdded = true;
                  };
                };
              };
            };

            if (haveSelectionRemoved)
              change(NULL, GC_CHANGE_SELECTION_REMOVE);

            if (haveSelectionAdded)
              change(NULL, GC_CHANGE_SELECTION_ADD);

            break;
          };
      };
      delete enumerator;
    };
    
    if (haveSelectionAdded || haveSelectionRemoved)
    {
      makeDirty();
      change(this, GC_CHANGE_SELECTION_CHANGE);
    }
    else
      canvas()->refresh();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Starts the rubber rectangle if none is active currently. Otherwise it does nothing.
 *
 * @param style Determines the visible style of the rubber rectangle.
 * @param coords The coordinates of the start point in view space.
 * @param removeSelection If true then the current selection will be cleared.
 */
void CFeedbackLayer::rubberRectStart(TRubberRectStyle style, const TVertex& coords, bool removeSelection)
{
  if ((FStates & GC_FBSTATE_RUBBERRECT) == 0)
  {
    FStates |= GC_FBSTATE_RUBBERRECT;

    if (removeSelection)
      clearSelection();

    FRubberRectBounds.upper = viewToLayer(coords);
    FRubberRectBounds.lower = FRubberRectBounds.upper;

    if (FRubberRectDisplayList == 0)
      FRubberRectDisplayList = glGenLists(1);
    glNewList(FRubberRectDisplayList, GL_COMPILE);
    glPushAttrib(GL_LINE_BIT);
    glDisable(GL_LINE_SMOOTH);

    switch (style)
    {
      case GC_RRSTYLE_SOLID_THIN:   // A simple black rectangle with a one pixel wide border.
        {
          GLubyte Color[4];

          if (!colorByName("Highlight", Color))
          {
            // If the system's highlight color could not be found then use a default one.
            Color[0] = 64;
            Color[1] = 64;
            Color[2] = 128;
          };

          Color[3] = 80;
          glColor4ubv(Color);
          glBegin(GL_POLYGON);
            glVertex3d(0, 0, 1);
            glVertex3d(1, 0, 1);
            glVertex3d(1, 1, 1);
            glVertex3d(0, 1, 1);
          glEnd();

          Color[3] = 255;
          glColor4ubv(Color);
          glLineWidth(1);
          glBegin(GL_LINE_LOOP);
            glVertex3d(0, 0, 1);
            glVertex3d(1, 0, 1);
            glVertex3d(1, 1, 1);
            glVertex3d(0, 1, 1);
          glEnd();

          break;
        };
      case GC_RRSTYLE_SOLID_THICK:  // A simple black rectangle with a 3 pixel wide border.
        {
          GLubyte Color[4];

          if (!colorByName("Highlight", Color))
          {
            // If the system's highlight color could not be found then use a default one.
            Color[0] = 64;
            Color[1] = 64;
            Color[2] = 128;
          };

          Color[3] = 80;
          glColor4ubv(Color);
          glBegin(GL_POLYGON);
            glVertex3d(0, 0, 1);
            glVertex3d(1, 0, 1);
            glVertex3d(1, 1, 1);
            glVertex3d(0, 1, 1);
          glEnd();

          Color[3] = 255;
          glColor4ubv(Color);
          glLineWidth(3);
          glBegin(GL_LINE_LOOP);
            glVertex3d(0, 0, 1);
            glVertex3d(1, 0, 1);
            glVertex3d(1, 1, 1);
            glVertex3d(0, 1, 1);
          glEnd();

          break;
        };
      case GC_RRSTYLE_DOTTED_THIN:  // A simple black rectangle with a one pixel wide dotted border.
        {
          glColor3f(0, 0, 0);
          glLineWidth(1);
          glEnable(GL_LINE_STIPPLE);
          glLineStipple(1, 0xFF);
          glBegin(GL_LINE_LOOP);
            glVertex3d(0, 0, 1);
            glVertex3d(1, 0, 1);
            glVertex3d(1, 1, 1);
            glVertex3d(0, 1, 1);
          glEnd();
          break;
        };
      case GC_RRSTYLE_DOTTED_THICK: // A simple black rectangle with a 3 pixel wide dotted border.
        {
          glColor3f(0, 0, 0);
          glLineWidth(3);
          glEnable(GL_LINE_STIPPLE);
          glLineStipple(1, 0xFF);
          glBegin(GL_LINE_LOOP);
            glVertex3d(0, 0, 1);
            glVertex3d(1, 0, 1);
            glVertex3d(1, 1, 1);
            glVertex3d(0, 1, 1);
          glEnd();
          break;
        };
      case GC_RRSTYLE_BLENDED_CHECKERBOARD: // A filled rectangle with a one pixel border and a translucent interior.
        {
          GLubyte Color[4];
          if (!colorByName("Highlight", Color))
          {
            // If the system's highlight color could not be found then use a default one.
            Color[0] = 64;
            Color[1] = 64;
            Color[2] = 128;
          };

          glLineWidth(1);
          glEnable(GL_LINE_SMOOTH);
          for (int Row = 0; Row < 10; Row++)
          {
            for (int Cell = 0; Cell < 10; Cell++)
            {
              if (((Row + Cell) & 1) == 0)
                Color[3] = 60;
              else
                Color[3] = 70;
              glColor4ubv(Color);
              glBegin(GL_POLYGON);
                glVertex3d(Cell * 0.1, Row * 0.1, 1);
                glVertex3d(Cell * 0.1, (Row + 1) * 0.1, 1);
                glVertex3d((Cell + 1) * 0.1, (Row + 1) * 0.1, 1);
                glVertex3d((Cell + 1) * 0.1, Row * 0.1, 1);
              glEnd();
            };
          };

          Color[3] = 200;
          glColor4ubv(Color);
          glBegin(GL_LINE_LOOP);
            glVertex3d(0, 0, 1);
            glVertex3d(1, 0, 1);
            glVertex3d(1, 1, 1);
            glVertex3d(0, 1, 1);
          glEnd();
          break;
        };
      case GC_RRSTYLE_BLENDED_DIAGONALS: // A filled rectangle with a one pixel border and a translucent interior.
        {
          GLubyte Color[4];
          if (!colorByName("Highlight", Color))
          {
            // If the system's highlight color could not be found then use a default one.
            Color[0] = 64;
            Color[1] = 64;
            Color[2] = 128;
          };

          glLineWidth(1);

          // Interior.
          int Steps = 50;
          double dStep = 1.0 / Steps;

          // left lower half.
          for (int x = 0; x < Steps; x++)
          {
            if (x % 2 == 0)
              Color[3] = 60;
            else
              Color[3] = 100;

            glColor4ubv(Color);
            glBegin(GL_POLYGON);
              glVertex3d(x * dStep, 0, 1);
              glVertex3d(0, x * dStep, 1);
              glVertex3d(0, (x + 1) * dStep, 1);
              glVertex3d((x + 1) * dStep, 0, 1);
            glEnd();
          };

          // right upper half.
          for (int y = 0; y < Steps; y++)
          {
            if (y % 2 == 0)
              Color[3] = 60;
            else
              Color[3] = 100;

            glColor4ubv(Color);
            glBegin(GL_POLYGON);
              glVertex3d(1, y * dStep, 1);
              glVertex3d(y * dStep, 1, 1);
              glVertex3d((y + 1) * dStep, 1, 1);
              glVertex3d(1, (y + 1) * dStep, 1);
            glEnd();
          };

          // Border.
          Color[3] = 200;
          glColor4ubv(Color);
          glBegin(GL_LINE_LOOP);
            glVertex3d(0, 0, 1);
            glVertex3d(1, 0, 1);
            glVertex3d(1, 1, 1);
            glVertex3d(0, 1, 1);
          glEnd();
          break;
        };
    };
    glPopAttrib();
    glEndList();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Stops the rubber rectangle if it is active currently. Does nothing if not.
 */
void CFeedbackLayer::rubberRectStop(void)
{
  if ((FStates & GC_FBSTATE_RUBBERRECT) != 0)
  {
    FStates &= ~GC_FBSTATE_RUBBERRECT;
    canvas()->refresh();
  };
}

//----------------- CConnectionInstanceListener --------------------------------------------------------------------------

void CConnectionLayer::CConnectionInstanceListener::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
  if (layer != NULL && reason == GC_CHANGE_CONNECTION_INSTANCE)
    layer->invalidateInstances((CConnectionInstance*) sender);
}

//----------------------------------------------------------------------------------------------------------------------

void CConnectionLayer::CConnectionInstanceListener::onDestroy(CGCBase* object)
{
  layer->removeInstance((CConnectionInstance*) object);
}

//----------------------------------------------------------------------------------------------------------------------

void CConnectionLayer::CConnectionInstanceListener::onError(CGCBase* sender, CGCBase* origin, const char* message)
{
  layer->error(origin, message);
}

//----------------- CConnectionLayer ------------------------------------------------------------------------------------

CConnectionLayer::CConnectionLayer(CGCView* view): CLayer("connections", view->canvas())
{
  _className = "CConnectionLayer";
  FView = view;
  FListener.layer = this;
  FHotConnection = NULL;
  FStates = 0;
  FCache = NULL;
  FCurrentFeedback = GC_FI_NONE;
}

//----------------------------------------------------------------------------------------------------------------------

CConnectionLayer::~CConnectionLayer(void)
{
  setDestroying();

  delete FCache;

  for (CConnectionInstanceList::iterator iterator = FConnections.begin(); iterator != FConnections.end(); ++iterator)
    delete *iterator;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Takes the given coordinates and tries to find a graphic element that was rendered at this position.
 * Positions are given in view space.
 *
 * @param point The point to check given in view space. If necessary convert window coordinates first by using windowToView.
 * @param singleHit If true then search for hits is stopped after the first one was found.
 * @return A hit result class is returned regardless of the actual number of hits. It must be freed by the caller.
 */
CHitResults* CConnectionLayer::getHitTestInfoAt(TVertex point, bool singleHit)
{
  CHitResults* result = new CHitResults();
                      
  updateCache();
  if (singleHit)
  {
    CGraphicElement* element = FCache->findElement(point);
    if (element != NULL)
      result->addHit(element);
  }
  else
    FCache->findElements(point, result);

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Marks all connection instances the are connected to the given endpoint as dirty, if they are not already.
 *
 * @param point The endpoint to iterate.
 */
void CConnectionLayer::invalidateEndPoint(CFigureInstance* point)
{
  CFigureConnectionList::iterator pointIterator = FConnectionPointList.find(point);
  for (CConnectionInstanceList::iterator iterator = pointIterator->second.begin(); iterator != pointIterator->second.end(); 
    ++iterator)
  {
    CConnectionInstance* instance = *iterator;
    if (!instance->dirty())
    {
      instance->makeDirty();
      if (instance->endPoint1() == point)
        invalidateEndPoint(instance->endPoint2());
      else
        invalidateEndPoint(instance->endPoint1());
    };
  };
}

//----------------------------------------------------------------------------------------------------------------------

// Binary predicates used to sort the direction lists in the next method.
// Return whether first element is less than the second

bool moreLeft(CConnectionInstance* element1, CConnectionInstance* element2)
{
  // Both connections share one end point. Compare the other.
  CFigureInstance* left;
  CFigureInstance* right;
  if (element1->endPoint1() == element2->endPoint1())
  {
    left = element1->endPoint2();
    right = element2->endPoint2();
  }
  else
  {
    left = element1->endPoint2();
    right = element2->endPoint1();
  };

  TBoundingBox bounds1 = left->bounds();
  float middle1 = (bounds1.lower.x + bounds1.upper.x) / 2;
  TBoundingBox bounds2 = right->bounds();
  float middle2 = (bounds2.lower.x + bounds2.upper.x) / 2;

  return middle1 < middle2;;
}

//----------------------------------------------------------------------------------------------------------------------

bool moreTop(CConnectionInstance* element1, CConnectionInstance* element2)
{
  // Both connections share one end point. Compare the other.
  CFigureInstance* left;
  CFigureInstance* right;
  if (element1->endPoint1() == element2->endPoint1())
  {
    left = element1->endPoint2();
    right = element2->endPoint2();
  }
  else
  {
    left = element1->endPoint2();
    right = element2->endPoint1();
  };

  TBoundingBox bounds1 = left->bounds();
  float middle1 = (bounds1.lower.y + bounds1.upper.y) / 2;
  TBoundingBox bounds2 = right->bounds();
  float middle2 = (bounds2.lower.y + bounds2.upper.y) / 2;

  return middle1 < middle2;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Computes the coordinates of all connection instances touching the given point.
 *
 * @param point The endpoint to validate.
 */
void CConnectionLayer::validateEndPoint(CFigureInstance* point)
{
  typedef vector<CConnectionInstance*> CConnectionList;

  CConnectionList northConnections;
  CConnectionList eastConnections;
  CConnectionList southConnections;
  CConnectionList westConnections;

  // In order to validate a connection instance both end points must be validated, that is, all
  // connections attached to these end points must be validated and the coordinates of the connection lines
  // must be computed.
  CFigureConnectionList::iterator pointIterator = FConnectionPointList.find(point);

  // Go through all connections attached to this point and determine the direction of them (i.e. where they are
  // attached to the point).
  for (CConnectionInstanceList::iterator iterator = pointIterator->second.begin(); iterator != pointIterator->second.end(); 
    ++iterator)
  {
    switch ((*iterator)->getDirection(point))
    {
      case GC_DIR_NORTH:
        {
          northConnections.push_back(*iterator);
          break;
        };
      case GC_DIR_EAST:
        {
          eastConnections.push_back(*iterator);
          break;
        };
      case GC_DIR_SOUTH:
        {
          southConnections.push_back(*iterator);
          break;
        };
      case GC_DIR_WEST:
        {
          westConnections.push_back(*iterator);
          break;
        };
    };
  };

  // Sort each direction list so connections do not overlap and compute for each direction the connection coordinates.
  int index = 1;
  sort(northConnections.begin(), northConnections.end(), moreLeft);
  for (CConnectionList::iterator iterator = northConnections.begin(); iterator != northConnections.end(); 
    ++iterator, ++index)
  {
    (*iterator)->computeCoordinates(point, GC_DIR_NORTH, (float) index / (northConnections.size() + 1));
  };

  index = 1;
  sort(eastConnections.begin(), eastConnections.end(), moreTop);
  for (CConnectionList ::iterator iterator = eastConnections.begin(); iterator != eastConnections.end(); 
    ++iterator, ++index)
  {
    (*iterator)->computeCoordinates(point, GC_DIR_EAST, (float) index / (eastConnections.size() + 1));
  };

  index = 1;
  sort(southConnections.begin(), southConnections.end(), moreLeft);
  for (CConnectionList ::iterator iterator = southConnections.begin(); iterator != southConnections.end(); 
    ++iterator, ++index)
  {
    (*iterator)->computeCoordinates(point, GC_DIR_SOUTH, (float) index / (southConnections.size() + 1));
  };

  index = 1;
  sort(westConnections.begin(), westConnections.end(), moreTop);
  for (CConnectionList ::iterator iterator = westConnections.begin(); iterator != westConnections.end(); 
    ++iterator, ++index)
  {
    (*iterator)->computeCoordinates(point, GC_DIR_WEST, (float) index / (westConnections.size() + 1));
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Renders all connections, which have been validate before in validateLayerContent.
 */
void CConnectionLayer::renderLayerContent(float currentZoom, TBoundingBox bounds)
{
  for (CConnectionInstanceList::iterator iterator = FConnections.begin(); iterator != FConnections.end(); ++iterator)
  {
    CConnectionInstance* instance = *iterator;
    if (instance->overlaps(bounds))
      instance->render(currentZoom, instance == FHotConnection);
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Checks if the layer's element cache must be created and validated.
 */
void CConnectionLayer::updateCache(void)
{
  if (FCache == NULL)
  {
    float width, height;
    FView->getWorkspace(&width, &height);
    FCache = new CElementLookupCache(width, height);
  };

  if (!FCache->isValid())
  {
    for (CConnectionInstanceList::iterator iterator = FConnections.begin(); iterator != FConnections.end(); ++iterator)
      FCache->addElement(*iterator);

    FCache->validate();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Computes the layout for all connection instances.
 * Note: the invalidation code (@see invalidateInstance) takes care that either all connection instances connected to an
 *       end point are dirty or none of them.
 */
void CConnectionLayer::validateLayerContent(void)
{
  // Iterate through all dirty connection instances and validates them.
  for (CConnectionInstanceList::iterator iterator = FConnections.begin(); iterator != FConnections.end(); ++iterator)
  {
    CConnectionInstance* instance = *iterator;
    if (instance->dirty())
    {
      validateEndPoint(instance->endPoint1());
      validateEndPoint(instance->endPoint2());
      instance->validate();
    };
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Creates a new connection instance.
 *
 * @param connection The connection for which the instance is to be created.
 * @param endPoint1 One end point of the instnace.
 * @param endPoint2 The other end point of the instance.
 */
CConnectionInstance* CConnectionLayer::createInstance(CConnection* connection, CFigureInstance* endPoint1, 
                                                      CFigureInstance* endPoint2)
{
  CConnectionInstance* result = new CConnectionInstance(connection, endPoint1, endPoint2);
  result->addListener(&FListener);
  FConnections.insert(result);
  FConnectionPointList[endPoint1].insert(result);
  FConnectionPointList[endPoint2].insert(result);
  makeDirty();

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Checks the given position and hit element for a detailed hit information.
 *
 * @param position The position to test against.
 * @return Info about the hit if any.
 */
TFeedbackInfo CConnectionLayer::getFeedbackInfo(TVertex position)
{
  TFeedbackInfo result = GC_FI_NONE;

  if (FStates == 0)
  {
    updateCache();

    CConnectionInstance* instance = (CConnectionInstance*) FCache->findElement(position);
    if (instance != NULL)
      result = instance->getFeedbackInfo(position.x, position.y);
  }
  else
    result = FCurrentFeedback;

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

bool CConnectionLayer::handleMouseDown(TMouseButton button, int modifiers, int windowX, int windowY, TVertex& viewCoords)
{
  bool result = false;

  updateCache();
  FLastHit = (CConnectionInstance*) FCache->findElement(viewCoords);
  if (FLastHit != NULL)
  {
    FLastWindowX = windowX;
    FLastWindowY = windowY;  
    FLastViewX = viewCoords.x;
    FLastViewY = viewCoords.y;
    TFeedbackInfo info = FLastHit->getFeedbackInfo(viewCoords.x, viewCoords.y);
    switch (info)
    {
    case GC_FI_CONNECTION:
      {
        // Select the connection.
        break;
      };
    case GC_FI_HLINE_MOVE:
      {
        FStates |= GC_CLAYER_STATE_HDRAG_PENDING;
        result = true;
        break;
      };
    case GC_FI_VLINE_MOVE:
      {
        FStates |= GC_CLAYER_STATE_VDRAG_PENDING;
        result = true;
        break;
      };
    };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

bool CConnectionLayer::handleMouseMove(int modifiers, int windowX, int windowY, TVertex& viewCoords)
{
  bool result = false;

  if (FStates == 0)
  {
    // No special action active so highlight the current hot node if hot tracking is enabled.
    updateCache();
    CGraphicElement* element = FCache->findElement(viewCoords);
    if (element != NULL)
    {
      if (element->classIs("CConnectionInstance") && element != FHotConnection)
      {
        FHotConnection = (CConnectionInstance*) element;
        canvas()->refresh();
      };
    }
    else
      if (FHotConnection != NULL)
      {
        FHotConnection = NULL;
        canvas()->refresh();
      };
  }
  else
  {
    // Check for center line dragging.
    if ((FStates & GC_CLAYER_STATE_HDRAG_PENDING) != 0)
    {
      // Drag is still waiting to start. Check if the user moved the mouse a bit.
      float threshold = 4; // Empirically found.
      if (abs(FLastWindowY - windowY) >= threshold)
      {
        FStates &= ~GC_CLAYER_STATE_HDRAG_PENDING;
        FStates |= GC_CLAYER_STATE_CENTER_DRAGGING;
        FCurrentFeedback = GC_FI_HLINE_MOVE;
      };
    }
    else
      if ((FStates & GC_CLAYER_STATE_VDRAG_PENDING) != 0)
      {
        // Drag is still waiting to start. Check if the user moved the mouse a bit.
        float threshold = 4; // Empirically found.
        if (abs(FLastWindowX - windowX) >= threshold)
        {
          FStates &= ~GC_CLAYER_STATE_VDRAG_PENDING;
          FStates |= GC_CLAYER_STATE_CENTER_DRAGGING;
          FCurrentFeedback = GC_FI_VLINE_MOVE;
        };
      };

    FLastViewX = viewCoords.x;
    FLastViewY = viewCoords.y;
    
    if ((FStates & GC_CLAYER_STATE_CENTER_DRAGGING) != 0)
    {
      FLastHit->moveCenterTo(viewCoords);
      FCache->invalidate();
      result = true;
    };
  };

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

bool CConnectionLayer::handleMouseUp(TMouseButton button, int modifiers, int windowX, int windowY, TVertex& viewCoords)
{
  bool result = true;

  FStates &= ~(GC_CLAYER_STATE_HDRAG_PENDING | GC_CLAYER_STATE_VDRAG_PENDING | GC_CLAYER_STATE_CENTER_DRAGGING);

  return result;
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Invalidates all connection instances that are connected to the end point of the given connection instance.
 *
 * @param The connection instance that has changed.
 */
void CConnectionLayer::invalidateInstances(CConnectionInstance* instance)
{
  FCache->invalidate();

  invalidateEndPoint(instance->endPoint1());
  invalidateEndPoint(instance->endPoint2());
  makeDirty();
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Removes the given instance from the internal list. The connection instance is not destroyed, though.
 *
 * @param instance The connection instance to be removed.
 */
void CConnectionLayer::removeInstance(CConnectionInstance* instance)
{
  if (!destroying())
  {
    FCache->invalidate();

    if (FHotConnection == instance)
      FHotConnection = NULL;

    FConnectionPointList[instance->endPoint1()].erase(instance);
    if (FConnectionPointList[instance->endPoint1()].size() == 0)
      FConnectionPointList.erase(instance->endPoint1());

    FConnectionPointList[instance->endPoint2()].erase(instance);
    if (FConnectionPointList[instance->endPoint2()].size() == 0)
      FConnectionPointList.erase(instance->endPoint2());

    FConnections.erase(instance);

    makeDirty();
  };
}

//----------------------------------------------------------------------------------------------------------------------

/**
 * Called by the owning view to let us know when the workspace size has changed. This directly influences the element cache
 * (if it uses a BSP tree).
 */
void CConnectionLayer::workspaceChanged(void)
{
  delete FCache;
  FCache = NULL;
}

//----------------------------------------------------------------------------------------------------------------------

