/* Copyright (C) 2005 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


// Notes:
//
// To add a view to the model:
// - create the view object
// - the view will be automatically added to the model
//
// To add a layer to the model:
// - create the layer object
// - setup coordinates and other stuff
// - add it to the layers list of the view
//
// To add an element to the model:
// - create the element
// - setup all its attributes
// - add it to the elementList of the rootLayer, it will be relocated to the correct
//   sublayer
//

#include "myx_grt_wb_controller.h"
#include "myx_grt_wb_public_interface.h"

#include <myx_gc_model.h>
#include <myx_gc_layer.h>


#define INTERNAL_ERROR(msg) g_warning("INTERNAL ERROR: %s", msg)

static inline MYX_GRT_VALUE *grtObjectFromValue(MYX_GRT *grt, MYX_GRT_VALUE *value)
{
  if (myx_grt_value_get_type(value) == MYX_STRING_VALUE)
    return myx_grt_reference_cache_lookup(grt, myx_grt_value_as_string(value));
  else
    return value;
}


static inline LayerController *layerFromValueObject(MYX_GRT_VALUE *value)
{
  if (myx_grt_value_get_type(value) == MYX_DICT_VALUE
      && strcmp2(myx_grt_dict_struct_get_name(value), "db.workbench.Layer")==0)
    return (LayerController*)myx_grt_value_bridge_data_object_get(value);
  return NULL;
}

static inline CanvasViewController *viewFromValueObject(MYX_GRT_VALUE *value)
{
  if (myx_grt_value_get_type(value) == MYX_DICT_VALUE
      && strcmp2(myx_grt_dict_struct_get_name(value), "db.workbench.View")==0)
    return (CanvasViewController*)myx_grt_value_bridge_data_object_get(value);
  return NULL;
}

static inline ElementController *elementFromValueObject(MYX_GRT *grt, MYX_GRT_VALUE *value)
{
  if (myx_grt_value_get_type(value) == MYX_DICT_VALUE)
  {
    MYX_GRT_STRUCT *gstruct= myx_grt_dict_struct_get(grt, value);
    if (!gstruct)
      return NULL;
      
    if (strcmp2(myx_grt_struct_get_parent_name(gstruct), "model.Element")==0)
      return (ElementController*)myx_grt_value_bridge_data_object_get(value);
  }
  else if (myx_grt_value_get_type(value) == MYX_STRING_VALUE)
  {
    return elementFromValueObject(grt, myx_grt_reference_cache_lookup(grt, myx_grt_value_as_string(value)));
  }
  return NULL;
}

static inline TableController *tableFromValueObject(MYX_GRT *grt, MYX_GRT_VALUE *value)
{
  if (myx_grt_value_get_type(value) == MYX_DICT_VALUE
      && strcmp2(myx_grt_dict_struct_get_name(value), "db.workbench.TableElement")==0)
    return (TableController*)myx_grt_value_bridge_data_object_get(value);
  return NULL;
}


static bool listContainsObject(MYX_GRT_VALUE *list, MYX_GRT_VALUE *value)
{
  int i, c;
  c= myx_grt_bridge_list_item_count(list);
  for (i= c-1; i >= 0; i--)
  {
    if (myx_grt_bridge_list_item_get(list, i, 0) == value)
      return true;
  }
  return false;
}


static int listIndexOfValue(MYX_GRT_VALUE *list, MYX_GRT_VALUE *value)
{
  int i, c;
  c= myx_grt_bridge_list_item_count(list);
  for (i= c-1; i >= 0; i--)
  {
    if (myx_grt_bridge_list_item_get(list, i, 0) == value)
      return i;
  }
  return -1;
}


static void removeFromList(MYX_GRT_VALUE *list, MYX_GRT_VALUE *value)
{
  int i, c;
  c= myx_grt_bridge_list_item_count(list);
  for (i= c-1; i >= 0; i--)
  {
    if (myx_grt_bridge_list_item_get(list, i, 0) == value)
    {
      myx_grt_bridge_list_item_del(list, i, 0);
      break;
    }
  }
}


RemakableObject::RemakableObject(WorkbenchController *wb)
: _wbench(wb), _pendingRemake(false)
{
}


void RemakableObject::setPendingRemake()
{
  if (!_pendingRemake)
  {
    _pendingRemake= true;
    _wbench->setPendingRemake(this);
  }
}


void RemakableObject::remakeDone()
{
  _pendingRemake= false;
}



BaseController::BaseController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: _wb(wb)
{
  _type= "base";
  _value= myx_grt_value_retain(value);
  _valueRetained= true;
}


BaseController::BaseController(WorkbenchController *wb)
: _wb(wb)
{
  _type= "base";
}


BaseController::~BaseController()
{
  g_message("destroying %s", _type);
  if (_valueRetained)
    myx_grt_value_release(_value);
}

string BaseController::description()
{
  return _type;
}


void BaseController::release()
{
  if (_valueRetained)
  {
    _valueRetained= false;
    myx_grt_value_release(_value);
  }
}


bool BaseController::updateFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  g_message("unhandled %s.updateFrom [%s]", _type, key);
  return false;
}


bool BaseController::updateToGrt(const char *key, MYX_GRT_VALUE *value)
{
//  g_message("unhandled %s.updateTo", _type);
  return false;
}


bool BaseController::valueFromGrt(const char *key, MYX_GRT_VALUE *value)
{
//  g_message("unhandled %s.valueFrom  %s", _type, key);
  if (strcmp2(key, "_id")==0)
  {
    _wb->rememberObject(this);
  }
  else
    return false;
  return true;
}

const char *BaseController::getStringValue(const char *key)
{
  MYX_GRT_VALUE *value= myx_grt_bridge_dict_item_get_value(_value, key, 0);
  if (value)
    return myx_grt_value_as_string(value);
  return NULL;
}

int BaseController::setStringValue(const char *key, const char *s)
{
  myx_grt_bridge_dict_item_set_value_from_string(_value, key, s, 0);

  return 0;
}

int BaseController::getIntValue(const char *key)
{
  MYX_GRT_VALUE *value= myx_grt_bridge_dict_item_get_value(_value, key, 0);
  if (value)
    return myx_grt_value_as_int(value);
  return 0;
}

double BaseController::getFloatValue(const char *key)
{
  MYX_GRT_VALUE *value= myx_grt_bridge_dict_item_get_value(_value, key, 0);
  if (value)
    return myx_grt_value_as_real(value);
  return 0.0;
}

MYX_GRT_VALUE *BaseController::getValue(const char *key)
{
  return myx_grt_bridge_dict_item_get_value(_value, key, 0);
}

bool BaseController::referenceIsSet(const char *key)
{
  const char *ref= getStringValue(key);
  if (!ref || !*ref)
    return false;
  return true;
}



void BaseController::setModel(ModelController *model)
{
  _ownerModel= model;
}


#define DECORATION_INDEX(many, mand) ((mand ? 1 : 0) | 1<<(many ? 1 : 0))

ModelController::ModelController(WorkbenchController *wb, CGenericCanvas *canvas, MYX_GRT_VALUE *value)
: BaseController(wb, value), _canvas(canvas)
{
  _type= "model";
  
  _connectionParts[DECORATION_INDEX(false,false)]= canvas->createFigure("db.Connection.End.Single.Optional", NULL);
  _connectionParts[DECORATION_INDEX(false,true)]= canvas->createFigure("db.Connection.End.Single.Mandatory", NULL);
  _connectionParts[DECORATION_INDEX(true,false)]= canvas->createFigure("db.Connection.End.Multiple.Optional", NULL);
  _connectionParts[DECORATION_INDEX(true,true)]= canvas->createFigure("db.Connection.End.Multiple.Mandatory", NULL);
  _connectionParts[4]= canvas->createFigure("db.Connection.Center", NULL);
  
  _connectionPartText[0]= canvas->createFigure("db.Connection.End1.Text", NULL);
  _connectionPartText[1]= canvas->createFigure("db.Connection.End2.Text", NULL);
  _connectionPartText[2]= canvas->createFigure("db.Connection.Center.Text", NULL);
}


ModelController::~ModelController()
{
  for (int i= 0; i < 5; i++)
    delete _connectionParts[i];
  for (int i= 0; i < 3; i++)
    delete _connectionPartText[i];
}


void ModelController::decorateConnection(CConnection *connection,
                                         bool sourceMandatory, bool sourceMany,
                                         bool destMandatory, bool destMany)
{
  TConnectionDecorations decoration;
  
  decoration.end1Decoration= _connectionParts[DECORATION_INDEX(sourceMany,sourceMandatory)];
  decoration.end1Label = _connectionPartText[0];
  
  decoration.centerDecoration = _connectionParts[4];
  decoration.centerLabel = _connectionPartText[2];
  
  decoration.end2Decoration = _connectionParts[DECORATION_INDEX(destMany,destMandatory)];
  decoration.end2Label = _connectionPartText[1];
  
  connection->decorations(decoration);
}


CanvasViewController* ModelController::currentView()
{
  CGCView *gcView= _canvas->currentView();
  
  return controllerForCanvasView(gcView);
}


void ModelController::addViewRef(CanvasViewController *view)
{
  myx_grt_bridge_list_item_insert(getValue("views"),
                                  -1,
                                  view->value(),
                                  0);
  
  view->release();
}


void ModelController::removeViewRef(CanvasViewController *view)
{
  removeFromList(getValue("views"), view->value());
}


void ModelController::deleteView(CanvasViewController *view)
{
  view->clear();
  
  removeViewRef(view);
  
  //XXX
}


ElementController *ModelController::controllerForElement(CGCBase *element)
{
  return (ElementController*)(BaseController*)element->userData();
}


void ModelController::changedCurrentView(CanvasViewController *view)
{
  const char *_id= view->getStringValue("_id");
  MYX_GRT_VALUE *value= getValue("currentView");
  if (_id)
    myx_grt_bridge_value_change_string(value, _id);
  else
  {
    char *new_id= myx_grt_get_guid();

    view->setStringValue("_id", new_id);
    myx_grt_bridge_value_change_string(value, new_id);
  }  
}


bool ModelController::updateFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  if (strcmp2(key, "views")==0)
  {
    MYX_GRT_VALUE *viewL= getValue("views");
    
    // check what happened to the views list
    for (map<string,CanvasViewController*>::const_iterator iter= _viewControllers.begin();
         iter != _viewControllers.end(); ++iter)
    {
      if (!listContainsObject(viewL, iter->second->value()))
      {
        CanvasViewController *viewCtl= iter->second;

        g_message("removed view %s", viewCtl->getStringValue("name"));
        delete viewCtl;
      }
    }
  }
  else if (strcmp2(key, "elements")==0)
  {
    
  }
  else
    return BaseController::updateFromGrt(key, value);
  return true;
}


bool ModelController::valueFromGrt(const char *key, MYX_GRT_VALUE *value)
{
//  g_message("value from grt, key=%s type=%s", key,
//            myx_get_value_type_as_string(myx_grt_value_get_type(value)));
  
  if (strcmp2(key, "views")==0)
  {
    INTERNAL_ERROR("Views are added to the model automatically.");
    removeFromList(getValue("views"), value);
  }
  else if (strcmp2(key, "elements")==0)
  {
    if (getValue("elements") != value)
    {
      INTERNAL_ERROR("Cannot add element directly to view");
      removeFromList(getValue("elements"), value);
    }
  } 
  else if (strcmp2(key, "currentView")==0)
  {
    CanvasViewController *view= viewFromValueObject(grtObjectFromValue(_wb->grt(), value));
    
    if (_canvas->currentView() != view->gcView())
    {
      _canvas->currentView(view->gcView());      
    }
  }
  else
  if (key)
  {
    g_message("MODEL update TO %s <%s>, %s", key,
              myx_get_value_type_as_string(myx_grt_value_get_type(value)),
              myx_grt_dict_struct_get_name(value));
  }
  return true;
}


void ModelController::rememberViewController(CanvasViewController *object)
{
  _viewControllers[object->getStringValue("name")]= object;
}


CanvasViewController *ModelController::controllerForCanvasView(CGCView *object)
{
  string name= object->property("name", 0);
  
  if (_viewControllers.find(name) != _viewControllers.end())
    return _viewControllers[name];
  
  return NULL;  
}

// ==============================================================================

CanvasViewController::CanvasViewController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: BaseController(wb, value), _gcView(0), _gcLayer(0), _rootLayer(0)
{
  _type= "view";
  
  
}


CanvasViewController::CanvasViewController(WorkbenchController *wb, CGCView *gcView)
: BaseController(wb), _gcView(gcView), _gcLayer(0), _rootLayer(0)
{
  _type= "view";
  string name= gcView->property("name", 0);

  _value= myx_grt_bridge_dict_new(_wb->grt(), "db.workbench.View", this);
  myx_grt_dict_init_obj(_wb->grt(), _value, name.c_str(), NULL, NULL);
}


CanvasViewController::~CanvasViewController()
{
  g_message("destroy view");
}


bool CanvasViewController::updateFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  if (strcmp2(key, "selection")==0)
  {
    g_message("selection changed");
  }
  else if (strcmp2(key, "layers")==0)
  {/*
    MYX_GRT_VALUE *layerL= getValue("layers");
    
    // check what happened to the layer list
    for (list<LayerController*>::iterator iter= _layers.begin();
         iter != _layers.end(); ++iter)
    {
      if (!listContainsObject(layerL, (*iter)->value()))
      {
        LayerController *layerCtl= *iter;
        //layerCtl->clear();
        g_message("removed sublayer %s", layerCtl->getStringValue("name"));        
        layerCtl->_parent= 0;
        //delete layerCtl;
        _layers.erase(iter);
        _wb->notifyLayer(layerCtl->value(), MYX_WBGC_REMOVED);
      }
    }
   */
  }
  else if (strcmp2(key, "elements")==0)
  {
  }
  else
    return BaseController::updateFromGrt(key, value);
  return true;
}


void CanvasViewController::postCreate()
{
  // this is called before the object is initialized, so create the root layer
  // here and add it to the layers list when it's initialized in valueFromGrt()
  
  MYX_GRT_VALUE *root_layer= myx_grt_dict_new_obj(_wb->grt(), "db.workbench.Layer", "root", "", 
                                                  _wb->model()->getStringValue("_id"));
  
  _rootLayer= layerFromValueObject(root_layer);
  
  _rootLayer->_parent= 0;
  _rootLayer->setView(this);
}


bool CanvasViewController::valueFromGrt(const char *key, MYX_GRT_VALUE *value)
{
//  g_message("VIEW valueFromGrt %s", key);
  if (strcmp2(key, "rootLayer")==0)
  {
    if (getValue("rootLayer") != value)
      INTERNAL_ERROR("Not possible to change root layer");
    else
      setStringValue("rootLayer", _rootLayer->getStringValue("_id"));
  }
  else if (strcmp2(key, "layers")==0)
  {
    if (getValue("layers") != value)
    {
      LayerController *lctl= layerFromValueObject(value);
      if (lctl)
      {
        if (_rootLayer != lctl)
        {
          // relocate must be called before setView() because it relies
          // on the frame not existing yet to know it's a new layer
          relocateLayerToContainerLayer(lctl);
          lctl->setView(this);
          _wb->notifyLayer(lctl->value(), MYX_WBGC_ADDED);
        }
      }
      else
      {
        INTERNAL_ERROR("Adding invalid layer object to view");
        removeFromList(getValue("layers"), value);
      }
    }
    else
    {
      // the layers list is being initialized, auto-add the root layer to it
      // and notify the app about it
      
      myx_grt_bridge_list_item_insert(getValue("layers"), -1, _rootLayer->value(), 0);
      
      _wb->notifyLayer(_rootLayer->value(), MYX_WBGC_ADDED);
    }
  }
  else if (strcmp2(key, "elements")==0)
  {  
    if (getValue("elements") != value)
    {
      ElementController *ectl= elementFromValueObject(_wb->grt(), value);
      
      if (ectl)
      {
        relocateElementToContainerLayer(ectl);      
      }
      else
      {
        INTERNAL_ERROR("Adding invalid element to elements list of view");
        removeFromList(getValue("elements"), value);
      }
    }
  }
  else if (strcmp2(key, "name")==0)
  {
    if (_ownerModel)
    {
      if (_ownerModel->canvas()->viewByName(getStringValue("name")))
      {
        INTERNAL_ERROR("Trying to create a view with a duplicate name");
      }
      else
      {
        if (!_gcView)
        {
          _ownerModel->rememberViewController(this);
          realizeView();
        }
      }
    }
  }
  else if (strcmp2(key, "selection")==0)
  {
    g_message("selection changed (added?)");
  }
  else
    return BaseController::valueFromGrt(key, value);
  return true;
}


void CanvasViewController::deleteLayer(LayerController *layer)
{
}


void CanvasViewController::clear()
{
}


void CanvasViewController::deleteElement(ElementController *elem)
{
  removeFromList(getValue("elements"), elem->value());
  elem->layer()->removeElementRef(elem);
  elem->release();
  
  elem->setPendingDestroy();
  g_message("DELETE! %i", _myx_grt_get_refcount(elem->value()));
}


void CanvasViewController::realizeView()
{
  CGenericCanvas *canvas= _ownerModel->canvas();
  
  g_message("GC:: creating view named %s", getStringValue("name"));
  
  canvas->beginUpdate();
  
  _gcView= canvas->createView(getStringValue("name"));
  _gcView->userData(this);
  _gcLayer= canvas->createLayer("Main", false);
  _gcLayer->userData(this);
  _gcView->addLayer(_gcLayer);
  
  canvas->endUpdate();
  
  _wb->notifyView(_gcView, MYX_WBGC_ADDED);
  _wb->notifyView(_gcView, MYX_WBGC_SWITCHED);
}


LayerController *CanvasViewController::findEnclosingLayer(const TBoundingBox &bounds)
{
  return _rootLayer->sublayerContainingBounds(bounds);
}


void CanvasViewController::relocateLayerToContainerLayer(LayerController *layer)
{
  TBoundingBox layerBounds= layer->grtBounds();

  // 0- if there's no frame created for the layer, it means it's a new layer.
  // put it inside the first sublayer that fully contains it, unless there's none
  // 1- look for sibling layers that are smaller than this one and fully contain it
  // 2- if there's no sibling layer, look for sibling layers of the parent layer 
  // (including parent itself)
  // 3- look for siblings of grandparent and so on
  // 4- if root layer is reached, then just relocate as child of root

  if (!layer->layerFrame())
  {
    bool found= false;
    // 0- this is a newcomer
    
    for (std::list<LayerController*>::reverse_iterator iter= _rootLayer->_layers.rbegin();
         iter != _rootLayer->_layers.rend(); ++iter)
    {
      if ((*iter)->containsBounds(layerBounds))
      {
        (*iter)->addSublayer(layer);
        found= true;
        return;
      }
    }
    if (!found)
      _rootLayer->addSublayer(layer);
  }
  else
  {
    g_message("LAYER RELOCATION NOT SUPPORTED");
  }
}


void CanvasViewController::relocateElementToContainerLayer(ElementController *element)
{
  TBoundingBox bounds= element->grtBounds();
  LayerController *layer;
  
  if (element->layer() && element->layer()->containsBounds(bounds))
    layer= element->layer();
  else
    layer= findEnclosingLayer(bounds);

  // if it's not totally in any layer, then leave it where it is
  if (!layer)
  {
    // unless it's not in it's old layer either
    if (!element->layer() || !element->layer()->containsBounds(bounds, false))
      element->setLayer(_rootLayer, false); // then move it to the root layer
    return;
  }
  
  // no changes
  if (layer == element->layer())
    return;
  
  // remove from current layer and add to new one
  element->setLayer(layer, false);
}


void CanvasViewController::addPendingLayerRelocation(LayerController *layer)
{
  // insert ordered by stacking
  _pendingLayerRelocations.push_back(layer);
}



void CanvasViewController::addPendingElementRelocation(ElementController *elem)
{
  // insert ordered by stacking
  _pendingElementRelocations.push_back(elem);
}


void CanvasViewController::relocatePendingObjects()
{
  // relocate layers
  for (std::list<LayerController*>::iterator iter= _pendingLayerRelocations.begin();
       iter != _pendingLayerRelocations.end(); ++iter)
  {
    relocateLayerToContainerLayer(*iter);
    (*iter)->clearPendingRelocation();    
  }
  _pendingLayerRelocations.clear();

  // relocate elements
  for (std::list<ElementController*>::iterator iter= _pendingElementRelocations.begin();
       iter != _pendingElementRelocations.end(); ++iter)
  {
    relocateElementToContainerLayer(*iter);
    (*iter)->clearPendingRelocation();
  }
  _pendingElementRelocations.clear();
}


void CanvasViewController::updateSelectionList()
{
  CFigureInstanceEnumerator *fenum= _gcView->getFigureInstanceEnumerator();
  MYX_GRT_VALUE *list= getValue("selection");

  while (fenum->hasNext())
  {
    CFigureInstance *instance= fenum->next();
    ElementController *element;
    MYX_GRT_VALUE *item;

    element= _ownerModel->controllerForElement(instance);
    if (!element)
    {
      continue;
    }
    item= element->value();

    if (instance->selected())
    {
      if (!listContainsObject(list, item))
        myx_grt_bridge_list_item_insert(list, -1, item, 0);
    }
    else
    {
      removeFromList(list, item);
    }
  }
}


// ==============================================================================


LayerController::LayerController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: BaseController(wb, value), RemakableObject(wb), _view(0), _parent(0), _gcFigure(0), _gcFigureInstance(0)
{
  _type= "layer";
}


LayerController::~LayerController()
{
}


void LayerController::setPendingRemake()
{
  RemakableObject::setPendingRemake();
    
  _wb->notifyLayer(_value, MYX_WBGC_MODIFIED);
}


void LayerController::remake()
{
  g_message("remake");
  createFrame();
}


TBoundingBox LayerController::grtBounds()
{
  TBoundingBox box;
  box.upper.x= (float) getFloatValue("left");
  box.upper.y= (float) getFloatValue("top");
  box.lower.x= box.upper.x + (float) getFloatValue("width");
  box.lower.y= box.upper.y + (float) getFloatValue("height");
  return box;
}


TBoundingBox LayerController::gcBounds()
{
  return _gcFigureInstance->bounds();
}


void LayerController::addSublayer(LayerController *layer)
{
  g_message("Adding %s as sublayer of %s", layer->getStringValue("name"),
            getStringValue("name"));
  
  if (layer->layerFrame())
  {
    // find the topmost layer and put the new one just above it
    if (_layers.begin() != _layers.end())
      _view->gcLayer()->bringToFront(layer->layerFrame(), (*_layers.rbegin())->layerFrame());
    else
      _view->gcLayer()->bringToFront(layer->layerFrame(), layerFrame());
  }
  myx_grt_bridge_list_item_insert(getValue("subLayers"), -1, 
                                  layer->getValue("_id"), 0);
  _layers.push_back(layer);
  layer->_parent= this;
    
  layer->commitStacking();
}


void LayerController::removeSublayer(LayerController *layer)
{
  layer->_parent= 0;
  _layers.erase(std::find(_layers.begin(), _layers.end(), layer));
  removeFromList(getValue("subLayers"), layer->getValue("_id"));
}


void LayerController::moveTo(float newX, float newY, bool dragging)
{
  double oldX, oldY;
  double dx, dy;
  MYX_GRT_VALUE *elements= getValue("elements");
  MYX_GRT *grt= _wb->grt();
  
  oldX= getFloatValue("left");
  oldY= getFloatValue("top");
  
  dx= newX - oldX;
  dy= newY - oldY;

  _ownerModel->canvas()->beginUpdate();
  
  // TODO: if the movement is the result of dragging, relocation of the contents
  // should be disabled
  
  // move the elements
  int c= myx_grt_bridge_list_item_count(elements);
  for (int i = 0; i < c; i++)
  {
    MYX_GRT_VALUE *item= myx_grt_bridge_list_item_get(elements, i, 0);
    ElementController *elem= elementFromValueObject(grt, item);

    if (elem && (!dragging || !elem->gcInstance()->selected()))
    {
      double x= elem->getFloatValue("left") + dx;
      double y= elem->getFloatValue("top") + dy;

      if (elem->gcInstance())
        elem->gcInstance()->translate((float)x, (float)y, 0, false);

      myx_grt_bridge_value_change_real(getValue("left"), x);
      myx_grt_bridge_value_change_real(getValue("top"), y);
    }
  }
  
  // move the sublayers
  for (std::list<LayerController*>::const_iterator iter= _layers.begin();
       iter != _layers.end(); ++iter)
  {
    double x= (*iter)->getFloatValue("left") + dx;
    double y= (*iter)->getFloatValue("top") + dy;
    if ((*iter)->_gcFigureInstance)
      (*iter)->_gcFigureInstance->translate((float)x, (float)y, 0, false);
    g_message("moving sublayer");
  }
  _ownerModel->canvas()->endUpdate();
  
  myx_grt_bridge_value_change_real(getValue("left"), newX);
  myx_grt_bridge_value_change_real(getValue("top"), newY);
}


void LayerController::onAction(CGCBase* sender, CGCBase* origin, TActionType actionType)
{
}


void LayerController::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
  if (reason == GC_CHANGE_FINSTANCE_PROPERTY)
  {
    TGCVariant gcValue = _gcFigureInstance->property("translation", 0);
    float newX, newY;
    
    newX= gcValue;
    gcValue = _gcFigureInstance->property("translation", 1);
    newY= gcValue;
    
    if (getFloatValue("left") != newX ||
        getFloatValue("top") != newY)
    {
      moveTo(newX, newY, true);
      if (!_moved)
      {
        _moved= true;
        _view->addPendingLayerRelocation(this);
      }
    }  
  }
  else if (reason == GC_CHANGE_FIGURE_RESIZE)
  {
    myx_grt_bridge_value_change_real(getValue("width"), _gcFigureInstance->width());
    
    myx_grt_bridge_value_change_real(getValue("height"), _gcFigureInstance->height());
  }
}


bool LayerController::updateFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  return BaseController::updateFromGrt(key, value);
}


bool LayerController::valueFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  if (strcmp2(key, "description")==0)
  {
  }
  else if (strcmp2(key, "name")==0)
  {    
    // apply name
    if (_gcFigure)
    {
      const char* name = getStringValue("name");
      _gcFigure->elementFromKey("/name")->setCaption(name != NULL ? name : "");
      _wb->canvas()->refresh();
      _wb->notifyLayer(_value, MYX_WBGC_MODIFIED);
    }
  }
  else if (strcmp2(key, "color")==0)
  {
    _wb->canvas()->beginUpdate();
    applyFrameProperties(ApplyColor);
    _wb->canvas()->endUpdate();
    _wb->notifyLayer(_value, MYX_WBGC_MODIFIED);
    _wb->canvas()->refresh();
  }
  else if (strcmp2(key, "top")==0)
  {
    applyFrameProperties(ApplyPos);
  }
  else if (strcmp2(key, "left")==0)
  {
    applyFrameProperties(ApplyPos);
  }
  else if (strcmp2(key, "width")==0)
  {
    applyFrameProperties(ApplySize);
  }
  else if (strcmp2(key, "height")==0)
  {
    applyFrameProperties(ApplySize);
  }
  else if (strcmp2(key, "subLayers")==0)
  {
    if (getValue("subLayers") != value)
    {
      INTERNAL_ERROR("Adding layer directly to subLayers list. Add to the layers list of the view instead");
      removeFromList(getValue("subLayers"), value);
    }
  }
  else if (strcmp2(key, "elements")==0)
  {
    if (getValue("elements") != value)
    {
      INTERNAL_ERROR("Adding element directly to a layer is not permitted. Add to the view instead");
      removeFromList(getValue("elements"), value);
    }
  }
  else
    return BaseController::valueFromGrt(key, value);
  
  return true;
}


LayerController *LayerController::topSublayer()
{
  if (_layers.empty())
    return 0;
  else
    return *_layers.rbegin();
}


void LayerController::addElementRef(ElementController *element)
{
  myx_grt_bridge_list_item_insert(getValue("elements"),
                                  -1,
                                  element->getValue("_id"),
                                  0);
}


void LayerController::removeElementRef(ElementController *element)
{
  removeFromList(getValue("elements"), element->getValue("_id"));
}


ElementController *LayerController::backmostElement(bool realizedOnly)
{
  MYX_GRT_VALUE *list= getValue("elements");
  int i, count;
  if ((count= myx_grt_bridge_list_item_count(list)) > 0)
  {
    for (i= 0; i < count; i++)
    {
      MYX_GRT_VALUE *item= myx_grt_bridge_list_item_get(list, i, 0);
      ElementController *elem= elementFromValueObject(_wb->grt(), item);
      
      if (!realizedOnly || elem->gcInstance())
        return elem;
    }
  }
  return 0;
}


ElementController *LayerController::realizedElementBelow(ElementController *elem)
{
  MYX_GRT_VALUE *list= getValue("elements");
  int i, count;

  if ((count= myx_grt_bridge_list_item_count(list)) > 0)
  {
    ElementController *prev= 0;
    
    for (i= 0; i < count; ++i)
    {
      MYX_GRT_VALUE *item= myx_grt_bridge_list_item_get(list, i, 0);
      ElementController *elemItem= elementFromValueObject(_wb->grt(), item);
      
      if (elemItem == elem)
        break;
      
      if (elemItem->gcInstance())
        prev= elemItem;
    }
    return prev;
  }  
  return 0;
}


ElementController *LayerController::frontmostElement(bool realizedOnly)
{
  MYX_GRT_VALUE *list= getValue("elements");
  int i, count;
  if ((count= myx_grt_bridge_list_item_count(list)) > 0)
  {
    for (i= count-1; i >= 0; i--)
    {
      MYX_GRT_VALUE *item= myx_grt_bridge_list_item_get(list, i, 0);
      ElementController *elem= elementFromValueObject(_wb->grt(), item);

      if (!realizedOnly || elem->gcInstance())
        return elem;
    }
  }  
  return 0;
}


void LayerController::raiseElement(ElementController *element, ElementController *above)
{
  // 1- put the element just above the one indicated in "above".
  // 2- if above is 0, then it will be put as the topmost element in this layer.
  // 3- if there's no elements in this layer, it will be put right above the topmost
  // sublayer frame (not its contents). 
  // 4- if there are no sublayers, then just put it right above the frame of this layer.
  // 5- if we're the root layer, put above the topmost sublayer
  
  g_message("raising element %s in layer", element->getStringValue("name"),
            getStringValue("name"));
  if (element->gcInstance() && element->layer()==this)
  {
    // the element to be put above must be in the same layer, otherwise
    // just put the element as the topmost in the current layer
    if (above && above->layer() != this)
      above= 0;
    
    // rearrange in canvas
    if (_view && _view->gcLayer())
    {
      CLayer *gcLayer= _view->gcLayer();

      // 1-
      if (above)
        gcLayer->bringToFront(element->gcInstance(), above->gcInstance());
      else
      {
        // 2- if above is 0, then make it the topmost element in its layer
        ElementController *elemUnder;

        // find the element directly under this
        elemUnder= frontmostElement(true);
        if (elemUnder == element)  // if we're already the frontmost element
          // find the element just below us
          elemUnder= realizedElementBelow(element);
        
        if (elemUnder)
          gcLayer->bringToFront(element->gcInstance(), elemUnder->gcInstance());
        else 
        {
          // 3- find the topmost of the sublayers in this layer
          if (_layers.begin() != _layers.end())
          {
            gcLayer->bringToFront(element->gcInstance(), 
                                (*_layers.rbegin())->_gcFigureInstance);
          }
          else
          {
            if (_gcFigureInstance) // 4- put right above the layer frame
              gcLayer->bringToFront(element->gcInstance(), _gcFigureInstance);
            else  // 5- send to front of all topmost sublayer
            {
              if (!_layers.empty())
                gcLayer->bringToFront(element->gcInstance(), 
                                     (*_layers.rbegin())->_gcFigureInstance);
              else
                gcLayer->sendToBack(element->gcInstance());
            }
          }
        }
      }
    }
    
    // rearrange in layer list
    if (_view)
    {
      myx_grt_value_retain(element->value());
      removeFromList(getValue("elements"), element->getValue("_id"));
      if (above)
      {
        int i= listIndexOfValue(getValue("elements"), above->value());
        myx_grt_bridge_list_item_insert(getValue("elements"), i+1,
                                        element->getValue("_id"), 0);
      }
      else
      {
        // add to the end of the list
        myx_grt_bridge_list_item_insert(getValue("elements"), -1,
                                        element->getValue("_id"), 0);
      }
      myx_grt_value_release(element->value());
    }
  }
}


void LayerController::lowerElement(ElementController *element, ElementController *under)
{
  if (element->layer()==this && (!under || under->layer()==this))
  {
    // rearrange in canvas
    if (_view && _view->gcLayer())
    {
      if (under)
        _view->gcLayer()->sendToBack(element->gcInstance(), under->gcInstance());
      else
      {
        LayerController *layer= element->layer();
        ElementController *under2= 0;
        while (layer)
        {
          under2= layer->backmostElement();
          if (under2)
            break;
          layer= layer->layerUnder();
        }
        
        if (!under2)
          _view->gcLayer()->bringToFront(element->gcInstance(), 0);
        else
          _view->gcLayer()->bringToFront(element->gcInstance(),
                                         under2->gcInstance());
      }
    }
    
    // rearrange in layer list
    if (_view)
    {
      myx_grt_value_retain(element->value());
      removeFromList(getValue("elements"), element->value());
      if (under)
      {
        int i= listIndexOfValue(getValue("elements"), under->value());
        myx_grt_bridge_list_item_insert(getValue("elements"), i,
                                        element->value(), 0);
      }
      else
      {
        // add to the start of the list
        myx_grt_bridge_list_item_insert(getValue("elements"), 0,
                                        element->value(), 0);
      }
      myx_grt_value_release(element->value());
    }    
  }  
}


LayerController *LayerController::layerAbove()
{
  if (_parent)
  {
    list<LayerController*>::const_iterator iter= find(_parent->_layers.begin(), _parent->_layers.end(), this);
    if (iter != _parent->_layers.end())
    {
      ++iter;
      return *iter;
    }
  }
  return 0;
}


LayerController *LayerController::layerUnder()
{
  if (_parent)
  {
    list<LayerController*>::const_iterator iter, prev;
    iter= _parent->_layers.begin();
    prev= _parent->_layers.end();
    while (iter != _parent->_layers.end())
    {
      if (*iter == this)
      {
        if (prev == _parent->_layers.end())
          return 0;
        else
          return *prev;
      }
      prev= iter;
      ++iter;
    }
  }
  return 0;
}


LayerController *LayerController::sublayerContainingBounds(const TBoundingBox &bounds)
{
  LayerController *layer;
  
  // traverse from topmost layer to bottom
  for (list<LayerController*>::reverse_iterator iter= _layers.rbegin();
       iter != _layers.rend(); ++iter)
  {
    layer= (*iter)->sublayerContainingBounds(bounds);
    if (layer)
      return layer;
    
    if ((*iter)->containsBounds(bounds))
      return *iter;
  }
  return 0;
}


void LayerController::setElementStyleColor(CFigureElement *element, const std::string &color)
{
  // set new color style for all child elements
  CElementList* elements= element->children();
  for (CElementList::iterator iterator= elements->begin();
    iterator != elements->end(); 
    ++iterator)
  {
    CFigureElement* child_element = *iterator;

    setElementStyleColor(child_element, color);
  }

  if (element->style())
  {
    // set new color style to the element itself
    std::wstring style_name= element->style()->name();
    std::wstring::size_type pos= style_name.find_last_of(L":");

    if (pos != std::wstring::npos)
    {
      std::wstring new_style_name= style_name.substr(0, pos + 1) + utf8ToUtf16(color);

      CGCModel *model= _wb->canvas()->getModel();
      if (model->styleExists(new_style_name))
      {
        CGCStyle *style= model->style(new_style_name);
        element->style(style);
      }
    }
  }
}


void LayerController::applyFrameProperties(int flags)
{
  _wb->canvas()->beginUpdate();
  
  if (_gcFigureInstance)
  {
    if (flags & ApplyColor)
    {
      // apply style
      const char *color= getStringValue("color");
      if (color && color[0])
      {
        setElementStyleColor(_gcFigure->content(), color);
      }
    }
    if (flags & ApplyPos)
    {
      // apply geometry
      _gcFigureInstance->translate((float)getFloatValue("left"),
                                   (float)getFloatValue("top"),
                                   0.0f);
    }
    if (flags & ApplySize)
    {
      _gcFigure->setSize((float)getFloatValue("width"),
                         (float)getFloatValue("height"));
    }
    
    if (flags & (ApplyPos|ApplySize))
      _view->relocateLayerToContainerLayer(this);
  }
  
  _wb->canvas()->endUpdate();
}


void LayerController::createFrame()
{
  CGCModel *model= _wb->canvas()->getModel();
  CFigureTemplate *layout = model->layout("db.workbench.LayerFrame", "full");

  if (layout != NULL)
  {
    CLayer *gcLayer= _view->gcLayer();
    _wb->canvas()->beginUpdate();

    _gcFigure = new CFigure(model, layout);

    CFigureElement* subElement = _gcFigure->elementFromKey("/name");
    const char* description = getStringValue("name");
    subElement->setCaption(description != NULL ? description : "Layer");
    
    _gcFigureInstance= gcLayer->createInstance(_gcFigure);
    _gcFigureInstance->addListener(this);

    g_message("CREATING FRAME");
    applyFrameProperties(ApplyAll);

    // put it in the right stacking order
    
    // 1- place above the sibling layer immediately under this one
    // 2- if there's no layer under, then put it right above the parent layer's frame
    // 3- if parent is root, then just put it in the bottom
    
    LayerController *under= layerUnder();
    if (under)
      gcLayer->bringToFront(_gcFigureInstance, under->layerFrame());
    else
    {
      if (_parent && _parent->layerFrame())
        gcLayer->bringToFront(_gcFigureInstance, _parent->layerFrame());
      else
        gcLayer->sendToBack(_gcFigureInstance);
    }
    
    _wb->canvas()->endUpdate();
    _wb->canvas()->refresh();
  }
}

void LayerController::setView(CanvasViewController *view)
{
  _view= view;
  
  if (_parent)
  {
    setPendingRemake();
//    // setup and bind a frame for the layer if this is not the root layer
//    createFrame();
  }
}


bool LayerController::containsBounds(const TBoundingBox &bounds,
                                     bool totally)
{
  double lx1, ly1, lx2, ly2;
  double fx1, fy1, fx2, fy2;
  
  lx1= getFloatValue("left");
  ly1= getFloatValue("top");
  lx2= lx1 + getFloatValue("width");
  ly2= ly1 + getFloatValue("height");
  
  fx1= bounds.upper.x;
  fy1= bounds.upper.y;
  fx2= bounds.lower.x;
  fy2= bounds.lower.y;
  
  if (totally)
  {
    if (fx1 >= lx1 && fy1 >= ly1 && fx2 <= lx2 && fy2 <= ly2)
      return true;
  }
  else
  {
    if (((fx1 >= lx1 && fx1 <= lx2) || (fx2 >= lx1 && fx2 <= lx2))
        && ((fy1 >= ly1 && fy1 <= ly2) || (fy2 >= ly1 && fy2 <= ly2)))
      return true;
  }

  return false;
}


void LayerController::commitStacking()
{
  int i, c;
  MYX_GRT_VALUE *list= getValue("elements");
  MYX_GRT *grt= _wb->grt();
  CLayer *gcLayer;
  
  if (!_view) return;
  
  gcLayer= _view->gcLayer();
  
  _ownerModel->canvas()->beginUpdate();
  
  if (_gcFigureInstance)
    gcLayer->bringToFront(_gcFigureInstance, 0);
  
  c= myx_grt_bridge_list_item_count(list);
  for (i = 0; i < c; i++)
  {
    MYX_GRT_VALUE *item= myx_grt_bridge_list_item_get(list, i, 0);
    ElementController *elem= elementFromValueObject(grt, item);
    if (elem && elem->gcInstance())
      gcLayer->bringToFront(_gcFigureInstance, 0);
  }
  
  for (std::list<LayerController*>::const_iterator iter= _layers.begin();
       iter != _layers.end(); ++iter)
  {
    (*iter)->commitStacking();
  }
  _ownerModel->canvas()->endUpdate();
}


void LayerController::clear()
{
}


// ==============================================================================

ElementController::ElementController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: BaseController(wb, value), RemakableObject(wb), _gcFigure(0), _gcFigureInstance(0), _layer(0)
{
  _pendingChange= false;
  _pendingDestroy= false;
  _type= "element";
  _resizable= true;
  _initialSizeSet= false;
}


TBoundingBox ElementController::grtBounds()
{
  TBoundingBox box;
  box.upper.x= (float) getFloatValue("left");
  box.upper.y= (float) getFloatValue("top");
  box.lower.x= box.upper.x + (float) getFloatValue("width");
  box.lower.y= box.upper.y + (float) getFloatValue("height");
  return box;
}


TBoundingBox ElementController::gcBounds()
{
  return _gcFigureInstance->bounds();
}


void ElementController::remake()
{
  if (!_pendingDestroy && !realizeConditionsMet())
  {
    INTERNAL_ERROR("trying to remake object in invalid state");
    return;
  }
   
  if (_layer)
  {
    bool firstTime= true;
    MYX_GRT_VALUE *object= objectData();
    if (!object)
      object= _value;

    // destroy the old table
    if (_gcFigureInstance)
    {
      firstTime= false;
      unrealizeFigure();
    }
    
    if (!_pendingDestroy)
    {
      // create new and add to canvas
      realizeFigure();
      
      if (firstTime && object != _value)
      {
        // this is the 1st time we're being called, setup the listeners
        setupListener(object);
      }
    }
  }
}


void ElementController::setPendingDestroy()
{
  _pendingDestroy= true;
  RemakableObject::setPendingRemake();
  
  _wb->notifyElement(_value, MYX_WBGC_REMOVED);
}


void ElementController::setPendingRemake()
{
  if (getIntValue("visible")!=0)
  {
    RemakableObject::setPendingRemake();
    
    // this will notify the app that something has changed and it should
    // call the pending flush callback whenever it can
    _wb->notifyElement(_value, MYX_WBGC_MODIFIED);
  }
}


int ElementController::listenerCallback(MYX_GRT *grt, MYX_GRT_VALUE *value, MYX_GRT_VALUE_CALLBACK_REASON reason, void *data)
{
  ElementController *self= (ElementController*)data;;

  if (!self->_pendingChange)
  {
    self->setPendingRemake();
    if (reason != MYX_GVCR_DELETE)
      self->setupListener(value);
  }
  return 0;
}


void ElementController::setupListener(MYX_GRT_VALUE *value)
{
  unsigned int i, count;
  MYX_GRT_VALUE *item;
  MYX_GRT_VALUE_TYPE type;
    
  switch (myx_grt_value_get_type(value))
  {
    case MYX_DICT_VALUE:
    {
      // duped listeners are checked in listener_add
      myx_grt_value_listener_add(_wb->grt(), value, this, &ElementController::listenerCallback);

      count= myx_grt_bridge_dict_item_count(value, 0);
      for (i= 0; i < count; i++)
      {
        item= myx_grt_bridge_dict_item_value_by_index(value, i, 0);

        type= myx_grt_value_get_type(item);
        if (type == MYX_DICT_VALUE || type == MYX_LIST_VALUE)
        {
          if (!myx_grt_value_listener_get(_wb->grt(), item, this, &ElementController::listenerCallback))
            setupListener(item);
        }
      }
      break;
    }
    case MYX_LIST_VALUE:
    {
      myx_grt_value_listener_add(_wb->grt(), value, this, &ElementController::listenerCallback);

      count= myx_grt_bridge_list_item_count(value);
      for (i= 0; i < count; i++)
      {
        item= myx_grt_bridge_list_item_get(value, i, 0);
        
        type= myx_grt_value_get_type(item);
        if (type == MYX_DICT_VALUE || type == MYX_LIST_VALUE)
        {
          if (!myx_grt_value_listener_get(_wb->grt(), item, this, &ElementController::listenerCallback))
            setupListener(item);
        }
      }
      break;
    }
  }
}


void ElementController::applyProperties()
{
  _gcFigureInstance->translate((float)getFloatValue("left"),
                            (float)getFloatValue("top"),
                            0);
  
  _gcFigure->setSize((float)getFloatValue("width"),
                     (float)getFloatValue("height"));

  TGCVariant var;
  var= getIntValue("expanded");
  if (_gcFigure->elementFromKey("/"))
    _gcFigure->elementFromKey("/")->property("expanded", 0, var);
}


void ElementController::addPartsFromDict(CFigureElement* element, MYX_GRT_VALUE *source, const string &path)
{
  const char *_id= myx_grt_dict_item_get_as_string(source, "_id");
  myx_grt_reference_cache_lookup(_wb->grt(), _id);
  
  unsigned int count = myx_grt_dict_item_count(source);
  for (unsigned int i = 0; i < count; ++i)
  {
    MYX_GRT_VALUE* child = myx_grt_dict_item_value_by_index(source, i);
    string localPath = path + "/" + myx_grt_dict_item_key_by_index(source, i);
    CFigureElement* subElement = element->figure()->elementFromKey(localPath.c_str());
    if (subElement != NULL)
    {
      switch (myx_grt_value_get_type(child))
      {
        case MYX_LIST_VALUE:
        {
          addPartsFromList(subElement, child, localPath);
          break;
        };
        case MYX_DICT_VALUE:
        {
          addPartsFromDict(subElement, child, localPath);
          break;
        };
        default:
        {
          const char* text = myx_grt_value_as_string(child);
          if (text != NULL)
            subElement->setCaption(text);
          break;
        };
      };
    };
  };
}


char *ElementController::formatColumn(MYX_GRT_VALUE *column)
{
  const char *name;
  const char *typeName;
  const char *typeParams;
  
  name= myx_grt_dict_name_item_as_string(column);
  typeName= myx_grt_dict_item_get_as_string(column, "datatypeName");
  typeParams= myx_grt_dict_item_get_as_string(column, "datatypeExplicitParams");
  
  if (typeParams && *typeParams)
    return g_strdup_printf("%s: %s(%s)", name, typeName, typeParams);
  else
    return g_strdup_printf("%s: %s", name, typeName);
}


void ElementController::addPartsFromList(CFigureElement* element, MYX_GRT_VALUE* source, const string &path)
{
  if (element != NULL)
  {
    // The sub element key that addresses the sub element in the figure element list is determined
    // by the type of the GRT values in the list.
    string localPath = path;
    bool isDbColumn = false;
    const char* elementType = myx_grt_list_content_get_struct_name(source);
    if (myx_grt_struct_check_is_or_inherits_from(_wb->grt(), elementType, "db.Column"))
    {
      localPath += "/column";
      isDbColumn = true;
    }
    else if (myx_grt_struct_check_is_or_inherits_from(_wb->grt(), elementType, "db.ForeignKey"))
      localPath += "/foreignKey";
    else if (myx_grt_struct_check_is_or_inherits_from(_wb->grt(), elementType, "db.Index"))
      localPath += "/index";
    else if (myx_grt_struct_check_is_or_inherits_from(_wb->grt(), elementType, "db.IndexColumn"))
      localPath += "/column";
    else if (myx_grt_struct_check_is_or_inherits_from(_wb->grt(), elementType, "db.Trigger"))
      localPath += "/trigger";

    element->userData(source);
    
    unsigned int count = myx_grt_list_item_count(source);
    switch (myx_grt_list_content_get_type(source))
    {
      case MYX_LIST_VALUE:
      {
        for (unsigned int i = 0; i < count; ++i)
        {
          MYX_GRT_VALUE* item = myx_grt_list_item_get(source, i);
          CFigureElement* child = element->addSubElement();
          addPartsFromList(child, item, localPath);
        };
        break;
      };
      case MYX_DICT_VALUE:
      {
        for (unsigned int i = 0; i < count; ++i)
        {
          MYX_GRT_VALUE* item = myx_grt_list_item_get(source, i);
          CFigureElement* child = element->addSubElement();

          // Format normal column name output differently.
          if (isDbColumn)
          {
            char* caption = formatColumn(item);
            if (caption)
            {
              string captionPath = localPath + "/name";
              CFigureElement* captionElement = element->figure()->elementFromKey(captionPath.c_str());
              if (captionElement != NULL)
                captionElement->setCaption(caption);
            };
            g_free(caption);
          }
          else
            addPartsFromDict(child, item, localPath);
        };
        break;
      };
      case MYX_STRING_VALUE:
      {
        // String list elements are usually reference ids.
        for (unsigned int i = 0; i < count; ++i)
        {
          MYX_GRT_VALUE* item = myx_grt_list_item_get(source, i);
//          CFigureElement* child = element->addSubElement();

          const char* caption = myx_grt_value_as_string(item);
          if (caption != NULL)
          {
            string captionPath = localPath + "/name";
            CFigureElement* captionElement = element->figure()->elementFromKey(captionPath.c_str());
            if (captionElement != NULL)
            {
              // This might be a normal property or a reference to another value. Check this.
              MYX_GRT_VALUE* target = myx_grt_reference_cache_lookup(_wb->grt(), caption);
              if (target != NULL)
              {
                const char* targetCaption = myx_grt_dict_name_item_as_string(target);
                if (targetCaption != NULL)
                  captionElement->setCaption(targetCaption);
                else
                  captionElement->setCaption(caption);
              };
            };
          };
        };
        break;
      };
      default:
      {
        // Not handled. Not sure what to do here.
        break;
      };
    };
  };
}


MYX_GRT_VALUE *ElementController::objectData()
{
  return NULL;
}


bool ElementController::realizeFigure()
{
  CGCModel* model = _wb->canvas()->getModel();
  MYX_GRT_STRUCT* grtType = myx_grt_dict_struct_get(_wb->grt(), _value);
  const char *layoutClass= "full";
  
  CFigureTemplate* layout;
  do
  {
    const char* typeName = myx_grt_struct_get_name(grtType);
    layout = model->layout(typeName, layoutClass);
    if (layout != NULL)
      break;

    const char* parentName = myx_grt_struct_get_parent_name(grtType);
    // Break out if we reached the top of the hierarchy.
    if (parentName == NULL)
      break;
    grtType = myx_grt_struct_get(_wb->grt(), parentName);

  } while (layout == NULL); 

  if (layout == NULL)
  {
    // If layout is NULL then we did not find a layout yet for the given type and class.
    // In this case we fall back to the icon class of the top element. This must
    // always be defined.
    const char* typeName = myx_grt_struct_get_name(grtType);
    layout = model->layout(typeName, "icon");
  }

  // If there is still no layout then we can't do anything.
  if (layout != NULL)
  {
    CLayer *gcLayer= _layer->view()->gcLayer();
    
    _gcFigure = new CFigure(model, layout);
    _gcFigure->userData((BaseController*)this);
    
    _gcFigure->beginUpdate();
    
    // setup the figure
    CFigureElement *content= _gcFigure->content();
    addPartsFromDict(content, objectData()?objectData():_value, "");
    
    // create an instance of the figure for the current view
    _gcFigureInstance= gcLayer->createInstance(_gcFigure);
    _gcFigureInstance->userData((BaseController*)this);
    _gcFigureInstance->addListener(this);
      
    // put the element in the right stacking order
    ElementController *elem= _layer->realizedElementBelow(this);
    if (elem)
      gcLayer->bringToFront(_gcFigureInstance, elem->gcInstance());
    else
    {
      if (_layer->layerFrame())
        gcLayer->bringToFront(_gcFigureInstance, _layer->layerFrame());
      else
        gcLayer->sendToBack(_gcFigureInstance);
    }

    if (!_initialSizeSet)
    {
      // get initial size
      myx_grt_bridge_value_change_real(getValue("width"), _gcFigureInstance->width());
    
      myx_grt_bridge_value_change_real(getValue("height"), _gcFigureInstance->height());
    }
    applyProperties();
    
    _gcFigure->endUpdate();
//    myx_grt_value_listener_add(_wb->grt(), _value, this, valueCallback);
    return true;
  }

  return false;
}


void ElementController::unrealizeFigure()
{
  _gcFigureInstance->removeListener(this);
  _layer->view()->gcLayer()->removeInstance(_gcFigureInstance);
  _gcFigureInstance= 0;
  
  // canvas will be notified automatically 
  delete _gcFigure;
  _gcFigure= 0;  
}


void ElementController::setLayer(LayerController *lctl, bool fromLayer)
{
  if (lctl != _layer)
  {
    if (!fromLayer)
    {
      if (_layer)
      {
        g_message("moving element %s from layer %s to %s",
                  getStringValue("name"), 
                  _layer->getStringValue("name"), lctl->getStringValue("name"));

        _layer->removeElementRef(this);
        _wb->notifyElement(_value, MYX_WBGC_REMOVED);
      }
      lctl->addElementRef(this);
    }
    myx_grt_bridge_value_change_string(getValue("layer"), lctl->getStringValue("_id"));

    _wb->notifyElement(_value, MYX_WBGC_ADDED);
    
    _layer= lctl;

    // always place at the top of the new layer
    _layer->raiseElement(this, 0);
  }
  
  if (!_gcFigureInstance && realizeConditionsMet())
    setPendingRemake();
}


bool ElementController::valueFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  if (strcmp2(key, "layer")==0)
  {
    LayerController *lctl= (LayerController*)_wb->objectForId(myx_grt_value_as_string(value));
    if (lctl)
    {
      setLayer(lctl, false);
      if (!_gcFigureInstance && realizeConditionsMet())
        setPendingRemake();
    }
  }
  else if (strcmp2(key, "top")==0)
  {
    if (_gcFigureInstance)
    {
      double y= myx_grt_value_as_real(value);
      double x= getFloatValue("left");
      _gcFigureInstance->translate((float)x, (float)y, 0);
    }
  }
  else if (strcmp2(key, "left")==0)
  {
    if (_gcFigureInstance)
    {
      double x= myx_grt_value_as_real(value);
      double y= getFloatValue("top");
      _gcFigureInstance->translate((float)x, (float)y, 0);
    }
  }
  else if (strcmp2(key, "width")==0)
  {
    if (getValue("width")!=value)
    {
      _initialSizeSet= true; 
      if (_gcFigureInstance)
      {
        if (_resizable)
        {
          applyProperties();
        }
        else
        {
          // read-only, undo change
          myx_grt_bridge_value_change_real(value, _gcFigureInstance->width());
        }
      }
    }
  }
  else if (strcmp2(key, "height")==0)
  {
    if (getValue("height")!=value)
    {
      _initialSizeSet= true;
      if (_gcFigureInstance)
      {
        if (_resizable)
        {
          applyProperties();
        }
        else
        {
          // read-only, undo change
          myx_grt_bridge_value_change_real(value, _gcFigureInstance->height());
        }
      }
    }
  }
  else if (strcmp2(key, "visible")==0)
  {
    if (!_gcFigureInstance && realizeConditionsMet())
      setPendingRemake();
  }
  else
    return BaseController::valueFromGrt(key, value);
  return true;
}


void ElementController::doEdit()
{
  MYX_GRT_ERROR error;
  MYX_GRT_VALUE *args;
  MYX_GRT_VALUE *result;
  
  args= myx_grt_list_new(MYX_ANY_VALUE, NULL);
  myx_grt_list_item_add(args, objectData());
  
  result= myx_grt_function_get_and_call(_wb->grt(),
                                        "Base","editObj",
                                        0, args, &error);
  if (result && myx_grt_dict_item_get_value(result, "error"))
  {
    g_message("error: %s", myx_grt_dict_item_get_as_string(result, "error"));
  }

  if (result) myx_grt_value_release(result);
  if (args) myx_grt_value_release(args);
}


void ElementController::doLock()
{
}


void ElementController::doMenu()
{
}


void ElementController::onAction(CGCBase* sender, CGCBase* origin, TActionType actionType)
{
  switch (actionType)
  {
    case GC_ACTION_APPLICATION+1: // Edit
      doEdit();
      break;
    case GC_ACTION_APPLICATION+2: // Lock
      doLock();
      break;
    case GC_ACTION_APPLICATION+3: // Menu
      doMenu();
      break;
  }
}


void ElementController::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
  if (reason == GC_CHANGE_FINSTANCE_PROPERTY)
  {
    TGCVariant gcValue = _gcFigureInstance->property("translation", 0);
    float newX, newY;
    
    newX= gcValue;
    gcValue = _gcFigureInstance->property("translation", 1);
    newY= gcValue;
    
    if (getFloatValue("left") != newX ||
        getFloatValue("top") != newY)
    {
      myx_grt_bridge_value_change_real(getValue("left"), newX);
      
      myx_grt_bridge_value_change_real(getValue("top"), newY);
    
      if (!_moved)
      {
        _moved= true;
        _layer->view()->addPendingElementRelocation(this);
      }
    }    
  }
  else if (reason == GC_CHANGE_FIGURE_RESIZE)
  {
    myx_grt_bridge_value_change_real(getValue("width"), _gcFigureInstance->width());
    
    myx_grt_bridge_value_change_real(getValue("height"), _gcFigureInstance->height());
  }
  else if (reason == GC_CHANGE_ELEMENT_PROPERTY)
  {
    onElementChange((CFigureElement*)origin);
  }
}


void ElementController::onElementChange(CFigureElement *element)
{
  TGCVariant gcValue = _gcFigure->content()->property("expanded", 0);
    
  if ((getIntValue("expanded")!=0) == (bool)gcValue)
  {
    myx_grt_bridge_value_change_int(getValue("expanded"), (int)gcValue);
  }      
}


bool ElementController::realizeConditionsMet()
{
  if (_layer && getIntValue("visible")!=0)
    return true;
  return false;
}

//---------------------------------------------------------------------------

TableController::TableController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: ElementController(wb, value)
{
  _type= "table";
  _resizable= false;
}


MYX_GRT_VALUE *TableController::objectData()
{
  const char *id= getStringValue("table");
  return myx_grt_reference_cache_lookup(_wb->grt(), id);
}


bool TableController::realizeConditionsMet()
{
  if (referenceIsSet("table") && ElementController::realizeConditionsMet())
    return true;
  return false;
}


void TableController::applyProperties()
{
  TGCVariant var;
  CFigureElement *elem;

  var= getIntValue("expanded");
  if ((elem = _gcFigure->elementFromId(L"/table-body")))
    elem->property("expanded", 0, var);
  
  var= getIntValue("columnsExpanded");
  if ((elem = _gcFigure->elementFromKey("/columns")))
    elem->property("expanded", 0, var);
  
  var= getIntValue("indicesExpanded");
  if ((elem = _gcFigure->elementFromKey("/indices")))
    elem->property("expanded", 0, var);
  
  var= getIntValue("foreignKeysExpanded");
  if ((elem = _gcFigure->elementFromKey("/foreignKeys")))
    elem->property("expanded", 0, var);
  
  var= getIntValue("triggersExpanded");
  if ((elem = _gcFigure->elementFromKey("/triggers")))
    elem->property("expanded", 0, var);
  
  const char *color= getStringValue("color");
  if (color && color[0])
  {
    _layer->setElementStyleColor(_gcFigure->content(), color);
  }      
  
  ElementController::applyProperties();
}


bool TableController::valueFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  bool handled= true;
  TGCVariant var;
  
  if (strcmp2(key, "table")==0)
  {
    if (myx_grt_value_get_type(value) == MYX_STRING_VALUE &&
        *myx_grt_value_as_string(value))
      
      if (!_gcFigureInstance && realizeConditionsMet())
        setPendingRemake();
  }  
  else if (strcmp2(key, "expanded")==0)
  {
    if (_gcFigureInstance)
    {
      var= myx_grt_value_as_int(value);
      _gcFigure->elementFromId(utf8ToUtf16("/table-body"))->property("expanded", 0, var);
    }
  }    
  else if (strcmp2(key, "columnsExpanded")==0)
  {
    if (_gcFigureInstance)
    {
      var= myx_grt_value_as_int(value);
      _gcFigure->elementFromKey("/columns")->property("expanded", 0, var);
    }
  }
  else if (strcmp2(key, "indicesExpanded")==0)
  {
    if (_gcFigureInstance)
    {
      var= myx_grt_value_as_int(value);
      _gcFigure->elementFromKey("/indices")->property("expanded", 0, var);
    }    
  }
  else if (strcmp2(key, "foreignKeysExpanded")==0)
  {
    if (_gcFigureInstance)
    {
      var= myx_grt_value_as_int(value);
      _gcFigure->elementFromKey("/foreignKeys")->property("expanded", 0, var);
    }
  }
  else if (strcmp2(key, "triggersExpanded")==0)
  {
    if (_gcFigureInstance)
    {
      var= myx_grt_value_as_int(value);
      _gcFigure->elementFromKey("/triggers")->property("expanded", 0, var);
    }
  }  
  else if (strcmp2(key, "color")==0)
  {
    _wb->canvas()->beginUpdate();
    if (_gcFigureInstance)
    {
      const char *color= getStringValue("color");
      if (color && color[0])
      {
        _layer->setElementStyleColor(_gcFigure->content(), color);
      }      
    }
    _wb->canvas()->endUpdate();
  }
  else
    handled= false;

  if (!handled)
    handled= ElementController::valueFromGrt(key, value);

  return handled;
}


void TableController::onElementChange(CFigureElement *element)
{
  if (_gcFigureInstance)
  {
    CFigureElement* element = _gcFigure->elementFromId(L"table-body");
    if (element != NULL)
      myx_grt_bridge_value_change_int(getValue("expanded"), element->property("expanded", 0));
    element = _gcFigure->elementFromKey("/columns");
    if (element != NULL)
      myx_grt_bridge_value_change_int(getValue("columnsExpanded"), element->property("expanded", 0));
    element = _gcFigure->elementFromKey("/indices");
    if (element != NULL)
      myx_grt_bridge_value_change_int(getValue("indicesExpanded"), element->property("expanded", 0));
    element = _gcFigure->elementFromKey("/foreignKeys");
    if (element != NULL)
      myx_grt_bridge_value_change_int(getValue("foreignKeysExpanded"), element->property("expanded", 0));
    element = _gcFigure->elementFromKey("/triggers");
    if (element != NULL)
      myx_grt_bridge_value_change_int(getValue("triggersExpanded"), element->property("expanded", 0));
  }
  
  //ElementController::onElementChange(element);
}


//----------------------------------------------------------------------------

ViewController::ViewController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: ElementController(wb, value)
{
  _type= "view";
  _resizable= false;
}

//----------------------------------------------------------------------------

RelationshipController::RelationshipController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: BaseController(wb, value), RemakableObject(wb), _gcConnection(0), _gcConnectionInstance(0)
{
  _type= "relationship";
}


bool RelationshipController::valueFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  if (strcmp2(key, "startTable")==0)
  {
  }
  else if (strcmp2(key, "endTable")==0)
  {
  }
  else if (strcmp2(key, "startMandatory")==0)
  {
  }
  else if (strcmp2(key, "endMandatory")==0)
  {
  }
  else if (strcmp2(key, "startMany")==0)
  {
  }
  else if (strcmp2(key, "endMany")==0)
  {
  }  
  else
    return BaseController::valueFromGrt(key, value);
  
  if (referenceIsSet("startTable") && referenceIsSet("endTable"))
  {
    if (!_gcConnection)
    {
      _wb->setPendingRemake(this);
    }
  }

  return true;
}


void RelationshipController::remake()
{
  TableController *startTable= (TableController*)_wb->objectForId(getStringValue("startTable"));
  TableController *endTable= (TableController*)_wb->objectForId(getStringValue("endTable"));

  if (!startTable || !endTable)
    return;
  
  LayerController *lctl= startTable->layer();
  if (!lctl)
    return;
  
  // destroy the old one
  
  // create new
  
  _gcConnection = _wb->canvas()->createConnection(startTable->gcFigure(), 
                                                  endTable->gcFigure());
  
  _ownerModel->decorateConnection(_gcConnection, 
                                  (getIntValue("startMandatory") == 1), (getIntValue("startMany") == 1),
                                  (getIntValue("endMandatory") == 1), (getIntValue("endMany") == 1));
  
  _gcConnectionInstance = lctl->view()->gcView()->createConnectionInstance(_gcConnection, 
                                                                           startTable->gcInstance(),
                                                                           endTable->gcInstance());
  _gcConnectionInstance->lineStyle(GC_CONNECTION_STYLE_SOLID);
}

//----------------------------------------------------------------------------

RoutineController::RoutineController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: ElementController(wb, value)
{
  _type= "routine";
  _resizable= false;
}

//----------------------------------------------------------------------------

NoteController::NoteController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: ElementController(wb, value)
{
  _type= "note";
}


bool NoteController::valueFromGrt(const char *key, MYX_GRT_VALUE *value)
{
  if (getValue("key")!=value)
    g_message("==> %s", key);
  if (strcmp2(key, "text")==0)
  {
    if (_gcFigureInstance)
      setPendingRemake();
  }
  else if (strcmp2(key, "color")==0)
  {
    if (_gcFigureInstance)
    {
      _wb->canvas()->beginUpdate();
      if (_gcFigureInstance)
      {
        const char *color= getStringValue("color");
        if (color && color[0])
        {
          _layer->setElementStyleColor(_gcFigure->content(), color);
        }      
      }
      _wb->canvas()->endUpdate();
    }
  }
  else
  {
    bool ret= ElementController::valueFromGrt(key, value);
    
    g_message("%i %p", getIntValue("visible"), _layer);
    
    if (!_gcFigureInstance && realizeConditionsMet())
      setPendingRemake();
    
    return ret;
  }
  return true;
}


void NoteController::applyProperties()
{
  _wb->canvas()->beginUpdate();
  if (_gcFigureInstance)
  {
    // apply style
    const char *color= getStringValue("color");
    if (color && color[0])
    {
      _layer->setElementStyleColor(_gcFigure->content(), color);
    }
  }
  _wb->canvas()->endUpdate();
  
  ElementController::applyProperties();
}

//----------------------------------------------------------------------------

ImageController::ImageController(WorkbenchController *wb, MYX_GRT_VALUE *value)
: ElementController(wb, value)
{
  _type= "image";
}


void ImageController::applyProperties()
{
  _wb->canvas()->beginUpdate();
  if (_gcFigureInstance)
  {
    const char *file= getStringValue("filename");
    
    
  }
  _wb->canvas()->endUpdate();
}



// ======================================================================


static struct {
  const char *name;
  BaseController *(*createController)(WorkbenchController *ctl, MYX_GRT_VALUE *value);
} WB_STRUCT_CONTROLLERS[]= {
  { "db.workbench.Model",         NULL },
  { "db.workbench.View",          &CanvasViewController::create },
  { "db.workbench.Layer",         &LayerController::create },
  { "db.workbench.Group",         NULL },
  { "db.workbench.Element",       &ElementController::create },
  { "db.workbench.TableElement",  &TableController::create },
  { "db.workbench.ViewElement",   &ViewController::create },
  { "db.workbench.RoutineElement",&RoutineController::create },
  { "db.workbench.NoteElement",   &NoteController::create },
  { "db.workbench.ImageElement",  &ImageController::create },
  { "db.workbench.Relationship",  &RelationshipController::create },
  { NULL, NULL }
};


WorkbenchController::WorkbenchController(MYX_GRT *grt)
: _grt(grt)
{  
//  myx_grt_value_bridge_dict_key_set(form_list, "forms");
}


WorkbenchController::~WorkbenchController()
{
  _canvas->removeListener(this);
}


void WorkbenchController::setCallbacks(void *data,
                  void (*view_changed)(MYX_GRT *Runtime, CGCView *view, MYX_WB_GC_CHANGE change, void *data),
                  void (*layer_changed)(MYX_GRT *Runtime, MYX_GRT_VALUE *layer, MYX_WB_GC_CHANGE change, void *data),
                  void (*element_changed)(MYX_GRT *Runtime, MYX_GRT_VALUE *element, MYX_WB_GC_CHANGE change, void *data))
{
  _callbackData= data;
  _viewChanged= view_changed;
  _layerChanged= layer_changed;
  _elementChanged= element_changed;
}



BaseController* WorkbenchController::objectForId(string oid)
{
  if (_objectTable.find(oid) != _objectTable.end())
    return _objectTable[oid];
  else
    return NULL;
}


void WorkbenchController::rememberObject(BaseController *object)
{
  _objectTable[myx_grt_dict_id_item_as_string(object->value())]= object;
}


void WorkbenchController::getObjectAtPoint(float x, float y, MYX_GRT_VALUE *&object, MYX_GRT_VALUE *&detail)
{
  CFigureInstance *figInstance;
  TVertex point;
  CHitResults *hits;
  
  object= 0;
  detail= 0;
  
  point.x= x;
  point.y= y;
  point.z= 0;
  point.w= 0;
  
  hits= _canvas->currentView()->getHitTestInfoAt(point, true);
  if (hits)
  {
    hits->reset();
    figInstance= (CFigureInstance*)hits->next();
    if (figInstance)
    {
      BaseController *controller= (BaseController*)figInstance->userData();
      
      if (strcmp(controller->type(), "layer")!=0)
      {
        ElementController *element= (ElementController*)controller;
        CFigureElement *figElement;
        
        object= element->value();
        
        figElement= element->gcInstance()->elementFromPoint(point);
        if (figElement)
          detail= (MYX_GRT_VALUE*)figElement->userData();
      }
      else
      {
        LayerController *layer= (LayerController*)controller;        
        
        object= layer->value();
      }
    }
  }
}


bool WorkbenchController::initObject(MYX_GRT_VALUE *object)
{
  MYX_GRT_STRUCT *gstruct= myx_grt_dict_struct_get(_grt, object);
  unsigned int i;
  BaseController *bctl= 0;

  if (strcmp2(gstruct->name, "db.workbench.Model")==0)
  {
    _modelCtl= new ModelController(this, _canvas, object);
    myx_grt_value_bridge_data_object_set(object, _modelCtl);

    return true;
  }
//  g_message("init object of type %s", gstruct->name);

  bctl= (BaseController*)myx_grt_value_bridge_data_object_get(object);
  if (!bctl)
  {
    for (i= 0; WB_STRUCT_CONTROLLERS[i].name; i++)
    {
      if (strcmp2(gstruct->name, WB_STRUCT_CONTROLLERS[i].name)==0)
      {
        if (WB_STRUCT_CONTROLLERS[i].createController)
          bctl= (*WB_STRUCT_CONTROLLERS[i].createController)(this, object);
        break;
      }
    }
    if (bctl)
    {      
      bctl->setModel(_modelCtl);
      myx_grt_value_bridge_data_object_set(object, bctl);
      
      bctl->postCreate();
      return true;
    }
  }
  else
    g_warning("OBJECT ALREADY HAS DATA SET");

  return false;
}


CanvasViewController *WorkbenchController::initObjectForView(CGCView *gcView)
{
  CanvasViewController *view;
  
  view= new CanvasViewController(this, gcView);
  
  view->setModel(_modelCtl);
      
  return view;
}


void WorkbenchController::setCanvas(CGenericCanvas *canvas)
{
  _canvas= canvas;
  _canvas->addListener(this);
}


void WorkbenchController::notifyView(CGCView *view, MYX_WB_GC_CHANGE change)
{
  if (_viewChanged)
    (*_viewChanged)(_grt, view, change, _callbackData);
}


void WorkbenchController::notifyLayer(MYX_GRT_VALUE *layer, MYX_WB_GC_CHANGE change)
{
  if (_layerChanged)
  {
    if (myx_grt_value_get_type(layer) == MYX_STRING_VALUE)
      layer= myx_grt_reference_cache_lookup(_grt, myx_grt_value_as_string(layer));
    
    (*_layerChanged)(_grt, layer, change, _callbackData);
  }
}


void WorkbenchController::notifyElement(MYX_GRT_VALUE *element, MYX_WB_GC_CHANGE change)
{
  if (_elementChanged)
  {
    if (myx_grt_value_get_type(element) == MYX_STRING_VALUE)
      element= myx_grt_reference_cache_lookup(_grt, myx_grt_value_as_string(element));

    (*_elementChanged)(_grt, element, change, _callbackData);
  }
}

void WorkbenchController::setPendingRemake(RemakableObject *element)
{
  _pendingRemakes.push_back(element);

  //added to make sure app does a refresh when it has time
  notifyElement(NULL, MYX_WBGC_MODIFIED);
}


void WorkbenchController::processPendingRemakes()
{
  _canvas->beginUpdate();
  for (list<RemakableObject*>::const_iterator iter= _pendingRemakes.begin();
       iter != _pendingRemakes.end(); ++iter)
  {
    (*iter)->remake();
    (*iter)->remakeDone();
  }
  _pendingRemakes.clear();
  _canvas->endUpdate();
}


void WorkbenchController::onChange(CGCBase* sender, CGCBase* origin, TGCChangeReason reason)
{
  CanvasViewController *view;
  //LayerController *layer;
  
  switch (reason)
  {
    case GC_CHANGE_CANVAS_ADD_VIEW:
    {
      // When Canvas reports a new view, it can mean:
      // - it's a view that we added ourselves, so we just need to add a reference
      // in the model list (in that case, _creatingObject will be true)
      // - it's a view added externally, so we have to create a GRT object for it
      // and add it to the model list
      view= _modelCtl->controllerForCanvasView(static_cast<CGCView*>(origin));
      if (!view)
      {
        view= initObjectForView(static_cast<CGCView*>(origin));
        _modelCtl->rememberViewController(view);
        _modelCtl->addViewRef(view);
      }
      else
      {
        _modelCtl->addViewRef(view);
        _modelCtl->changedCurrentView(view);
      }
      break;
    }
    case GC_CHANGE_CANVAS_REMOVE_VIEW:
      g_message("REMOVE VIEW");
      break;
      
    case GC_CHANGE_CANVAS_SWITCH_VIEW:
      view= _modelCtl->controllerForCanvasView(static_cast<CGCView*>(origin));
      if (view)
      {
        _modelCtl->changedCurrentView(view);
      }
      break;

    case GC_CHANGE_SELECTION_ADD:
    case GC_CHANGE_SELECTION_REMOVE:
    case GC_CHANGE_SELECTION_CLEAR:
    case GC_CHANGE_SELECTION_CHANGE:
      _modelCtl->currentView()->updateSelectionList();
      break;
  }
}




// -----------------------------------------------------------------------------
extern "C" {
  
MYX_GRT_VALUE * wb_bringToFront(MYX_GRT_VALUE *param, void *data)
{
  WorkbenchController *controller= (WorkbenchController*)data;
  MYX_GRT_VALUE *figure, *relative;
  ElementController *figureE, *relativeE;
  
  if (myx_grt_list_item_count(param) == 1)
  {
    figure= myx_grt_list_item_get(param, 0);
    relative= NULL;
  }
  else if (myx_grt_list_item_count(param) == 2)
  {
    figure= myx_grt_list_item_get(param, 0);
    relative= myx_grt_list_item_get(param, 1);
  }
  else
    return make_return_value_error("invalid # of arguments to bringToFront({element, relativeTo})", "");

  figureE= elementFromValueObject(controller->grt(), figure);
  if (relative)
    relativeE= elementFromValueObject(controller->grt(), relative);
  else
    relativeE= 0;
  
  if (!figureE)
    return make_return_value_error("invalid argument to bringToFront(): invalid element object", "");

  if (relativeE && relativeE->layer() != figureE->layer())
    return make_return_value_error("invalid argument to bringToFront(): elements are in different layers", "");

  if (figureE->layer())
    figureE->layer()->raiseElement(figureE, relativeE);
  else
    return make_return_value_error("invalid argument to bringToFront(): element is not in a layer", "");
  
  return make_return_value(myx_grt_value_from_int(1));
}


MYX_GRT_VALUE * wb_sendToBack(MYX_GRT_VALUE *param, void *data)
{
  WorkbenchController *controller= (WorkbenchController*)data;
  MYX_GRT_VALUE *figure, *relative;
  ElementController *figureE, *relativeE;
  
  if (myx_grt_list_item_count(param) == 1)
  {
    figure= myx_grt_list_item_get(param, 0);
    relative= NULL;
  }
  else if (myx_grt_list_item_count(param) == 2)
  {
    figure= myx_grt_list_item_get(param, 0);
    relative= myx_grt_list_item_get(param, 1);
  }
  else
    return make_return_value_error("invalid # of arguments to sendToBack({element, relativeTo})", "");
  
  figureE= elementFromValueObject(controller->grt(), figure);
  if (relative)
    relativeE= elementFromValueObject(controller->grt(), relative);
  else
    relativeE= 0;
  
  if (!figureE)
    return make_return_value_error("invalid argument to sendToBack(): invalid element object", "");
  
  if (relativeE && relativeE->layer() != figureE->layer())
    return make_return_value_error("invalid argument to sendToBack(): elements are in different layers", "");
  
  if (figureE->layer())
    figureE->layer()->lowerElement(figureE, relativeE);
  else
    return make_return_value_error("invalid argument to sendToBack(): element is not in a layer", "");
  return make_return_value(myx_grt_value_from_int(1));
}


MYX_GRT_VALUE * wb_objectAtPoint(MYX_GRT_VALUE *param, void *data)
{
  WorkbenchController *controller= (WorkbenchController*)data;
  MYX_GRT_VALUE *figure= NULL, *detail= NULL;  
  float x, y;
  
  if (myx_grt_list_item_count(param) == 2)
  {
    x= (float) myx_grt_value_as_real(myx_grt_list_item_get(param, 0));
    y= (float) myx_grt_value_as_real(myx_grt_list_item_get(param, 1));
  }
  else
    return make_return_value_error("invalid # of arguments to objectAtPoint({x, y})", "");
  
  controller->getObjectAtPoint(x, y, figure, detail);
  
  if (figure)
  {
    MYX_GRT_VALUE *result;

    result= myx_grt_dict_new(NULL, NULL);
    myx_grt_dict_item_set_value(result, "figure", figure);
    myx_grt_dict_item_set_value(result, "detail", detail);

    return make_return_value(result);
  }
  return make_return_value_error("No object at given position.", NULL);
}


MYX_GRT_VALUE *wb_deleteElement(MYX_GRT_VALUE *param, void *data)
{
  WorkbenchController *wb= (WorkbenchController*)data;
  MYX_GRT_VALUE *elemValue= myx_grt_bridge_list_item_get(param, 0, 0);
  ElementController *elem;

  elem= elementFromValueObject(wb->grt(), elemValue);
  if (elem)
    elem->layer()->view()->deleteElement(elem);
  
  return make_return_value(myx_grt_value_from_int(1));
}


MYX_GRT_VALUE *wb_deleteView(MYX_GRT_VALUE *param, void *data)
{
  WorkbenchController *wb= (WorkbenchController*)data;
  MYX_GRT_VALUE *viewValue= myx_grt_bridge_list_item_get(param, 0, 0);
  CanvasViewController *view;
  
  view= viewFromValueObject(viewValue);
  if (view)
    wb->model()->deleteView(view);
  
  return make_return_value(myx_grt_value_from_int(1));
}


MYX_GRT_VALUE *wb_deleteLayer(MYX_GRT_VALUE *param, void *data)
{
  MYX_GRT_VALUE *layerValue= myx_grt_bridge_list_item_get(param, 0, 0);
  LayerController *layer;
  
  layer= layerFromValueObject(layerValue);
  if (layer)
    layer->view()->deleteLayer(layer);
  
  return make_return_value(myx_grt_value_from_int(1));
}


MYX_GRT_VALUE * wb_selectElement(MYX_GRT_VALUE *param, void *data)
{
  WorkbenchController *wb= (WorkbenchController*)data;
  MYX_GRT_VALUE *elemValue= myx_grt_bridge_list_item_get(param, 0, 0);
  ElementController *elem;
  
  elem= elementFromValueObject(wb->grt(), elemValue);
  if (elem)
    elem->layer()->view()->gcView()->addToSelection(elem->gcInstance());
  
  return make_return_value(myx_grt_value_from_int(1));
}


MYX_GRT_VALUE * wb_unselectElement(MYX_GRT_VALUE *param, void *data)
{
  WorkbenchController *wb= (WorkbenchController*)data;
  MYX_GRT_VALUE *elemValue= myx_grt_bridge_list_item_get(param, 0, 0);
  ElementController *elem;
  
  elem= elementFromValueObject(wb->grt(), elemValue);
  if (elem)
    elem->layer()->view()->gcView()->removeFromSelection(elem->gcInstance());
  
  return make_return_value(myx_grt_value_from_int(1));
}


};
